import unittest
import json

from policy_sentry.command.write_policy import write_policy_with_template


valid_cfg_for_crud = {
    "mode": "crud",
    "name": "RoleNameWithCRUD",
    "read": [
        "arn:aws:s3:::example-org-sbx-vmimport",
        "arn:aws:s3:::example-kinnaird",
        "arn:aws:ssm:us-east-1:123456789012:parameter/test",
        "arn:aws:ssm:us-east-1:123456789012:parameter/test2",
        "arn:aws:kms:us-east-1:123456789012:key/123456",
        "arn:aws:s3:::job/jobid",
        "arn:aws:s3:::example-org-sbx-vmimport/stuff",
    ],
    "write": [
        "arn:aws:s3:::example-org-s3-access-logs",
        "arn:aws:s3:::example-org-sbx-vmimport/stuff",
        "arn:aws:secretsmanager:us-east-1:12345:secret:mysecret",
        "arn:aws:kms:us-east-1:123456789012:key/123456",
    ],
    "list": [
        "arn:aws:s3:::example-org-flow-logs",
        "arn:aws:s3:::example-org-sbx-vmimport/stuff",
    ],
    "tagging": ["arn:aws:ssm:us-east-1:123456789012:parameter/test"],
    "permissions-management": ["arn:aws:s3:::example-org-s3-access-logs"],
}

valid_cfg_for_actions = {
    "mode": "actions",
    "name": "RoleNameWithActions",
    "actions": [
        "kms:CreateGrant",
        "kms:CreateCustomKeyStore",
        "ec2:AuthorizeSecurityGroupEgress",
        "ec2:AuthorizeSecurityGroupIngress",
    ],
}


class YamlValidationOverallTestCase(unittest.TestCase):
    def test_allow_missing_name(self):
        """
        test_actions_missing_name: write-policy when the YAML file is missing a name
        :return:
        """
        cfg_with_missing_name = {
            "mode": "actions",
            "actions": [
                "kms:CreateGrant",
                "kms:CreateCustomKeyStore",
                "ec2:AuthorizeSecurityGroupEgress",
                "ec2:AuthorizeSecurityGroupIngress",
            ],
        }
        #  This should NOT raise an exception so leaving it as-is.
        policy = write_policy_with_template(cfg_with_missing_name)

    def test_allow_missing_arn(self):
        """
        test_actions_missing_arn: write-policy actions command when YAML file block is missing an ARN
        :return:
        """
        cfg_with_missing_actions = {
            "mode": "actions",
            "name": "RoleNameWithActions",
            "actions": [
                "kms:CreateGrant",
                "kms:CreateCustomKeyStore",
                "ec2:AuthorizeSecurityGroupEgress",
                "ec2:AuthorizeSecurityGroupIngress",
            ],
        }
        #  This should NOT raise an exception so leaving it as-is.
        policy = write_policy_with_template(cfg_with_missing_actions)


class YamlValidationCrudTestCase(unittest.TestCase):
    def test_allow_missing_access_level_categories_in_cfg(self):
        """
        test_allow_missing_access_level_categories_in_cfg: write-policy when the YAML file is missing access level categories. It should write a policy regardless.
        """

        crud_file_input = {
            "mode": "crud",
            "name": "RoleNameWithCRUD",
            "read": [
                "arn:aws:ssm:us-east-1:123456789012:parameter/test",
            ],
            "write": [
                "arn:aws:ssm:us-east-1:123456789012:parameter/test",
            ],
            "list": [
                "arn:aws:ssm:us-east-1:123456789012:parameter/test",
            ],
        }
        result = write_policy_with_template(crud_file_input)
        print(json.dumps(result, indent=4))
        self.maxDiff = None

    def test_empty_strings_in_access_level_categories(self):
        """
        test_empty_strings_in_access_level_categories: If the content of a list is an empty string, it should NOT sysexit
        :return:
        """
        crud_file_input = {
            "mode": "crud",
            "name": "RoleNameWithCRUD",
            "read": [
                "arn:aws:ssm:us-east-1:123456789012:parameter/test",
            ],
            "write": [
                "arn:aws:ssm:us-east-1:123456789012:parameter/test",
            ],
            "list": [
                "arn:aws:ssm:us-east-1:123456789012:parameter/test",
            ],
            "tagging": [""],
            "permissions-management": [""],
        }
        desired_output = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "SsmReadParameter",
                    "Effect": "Allow",
                    "Action": [
                        "ssm:GetParameter",
                        "ssm:GetParameterHistory",
                        "ssm:GetParameters",
                        "ssm:GetParametersByPath",
                    ],
                    "Resource": ["arn:aws:ssm:us-east-1:123456789012:parameter/test"],
                },
                {
                    "Sid": "SsmWriteParameter",
                    "Effect": "Allow",
                    "Action": [
                        "ssm:DeleteParameter",
                        "ssm:DeleteParameters",
                        "ssm:LabelParameterVersion",
                        "ssm:PutParameter",
                    ],
                    "Resource": ["arn:aws:ssm:us-east-1:123456789012:parameter/test"],
                },
                {
                    "Sid": "SsmListParameter",
                    "Effect": "Allow",
                    "Action": ["ssm:ListTagsForResource"],
                    "Resource": ["arn:aws:ssm:us-east-1:123456789012:parameter/test"],
                },
            ],
        }
        # with self.assertRaises(Exception):
        result = write_policy_with_template(crud_file_input)
        print(json.dumps(result, indent=4))
        # self.assertDictEqual(desired_output, result)
        # Future proofing tests - just checking if the ones we expect in the past are available in the future, even if we add on new ones.
        expected_sids = ["SsmReadParameter", "SsmWriteParameter", "SsmListParameter"]
        expected_write_actions = [
            "ssm:DeleteParameter",
            "ssm:DeleteParameters",
            "ssm:LabelParameterVersion",
            "ssm:PutParameter",
        ]
        expected_read_actions = [
            "ssm:GetParameter",
            "ssm:GetParameterHistory",
            "ssm:GetParameters",
            "ssm:GetParametersByPath",
        ]
        expected_list_actions = ["ssm:ListTagsForResource"]
        for statement in result["Statement"]:
            self.assertTrue(statement["Sid"] in expected_sids)
            if statement["Sid"] == "SsmWriteParameter":
                for action in expected_write_actions:
                    self.assertTrue(action in statement["Action"])
            if statement["Sid"] == "SsmReadParameter":
                for action in expected_read_actions:
                    self.assertTrue(action in statement["Action"])
            if statement["Sid"] == "SsmListParameter":
                for action in expected_list_actions:
                    self.assertTrue(action in statement["Action"])


class YamlValidationActionsTestCase(unittest.TestCase):
    def test_actions_missing_actions(self):
        """
        test_actions_missing_actions: write-policy actions if the actions block is missing
        :return:
        """
        cfg_with_missing_actions = {
            "mode": "actions",
            "name": "RoleNameWithActions",
        }
        with self.assertRaises(Exception):
            policy = write_policy_with_template(cfg_with_missing_actions)



import unittest
from policy_sentry.writing.template import create_actions_template, create_crud_template


class TemplateTestCase(unittest.TestCase):
    def test_actions_template(self):
        desired_msg = """mode: actions
name: ''
actions:
- ''
"""
        actions_template = create_actions_template()
        self.assertEqual(desired_msg, actions_template)

    def test_crud_template(self):
        desired_msg = """mode: crud
name: ''
# Specify resource ARNs
read:
- ''
write:
- ''
list:
- ''
tagging:
- ''
permissions-management:
- ''
# Actions that do not support resource constraints
wildcard-only:
  single-actions: # standalone actions
  - ''
  # Service-wide - like 's3' or 'ec2'
  service-read:
  - ''
  service-write:
  - ''
  service-list:
  - ''
  service-tagging:
  - ''
  service-permissions-management:
  - ''
# Skip resource constraint requirements by listing actions here.
skip-resource-constraints:
- ''
# Exclude actions from the output by specifying them here. Accepts wildcards, like kms:Delete*
exclude-actions:
- ''
# If this policy needs to include an AssumeRole action
sts:
  assume-role:
    - ''
  assume-role-with-saml:
    - ''
  assume-role-with-web-identity:
    - ''
"""
        crud_template = create_crud_template()
        self.maxDiff = None
        self.assertEqual(desired_msg, crud_template)



import unittest

from policy_sentry.querying.services import get_services_data


class QueryServicesTestCase(unittest.TestCase):
    def test_get_services_data(self):
        # when
        results = get_services_data()

        # then
        self.assertGreater(len(results), 400)  # in 07/24 it was 405

        # both should be a non-empty string
        self.assertTrue(results[0]["prefix"])
        self.assertTrue(results[0]["service_name"])



"""IAM Database queries that are not specific to either the Actions, ARNs, or Condition Keys tables."""

from __future__ import annotations

import functools
import logging

from policy_sentry.shared.constants import (
    POLICY_SENTRY_SCHEMA_VERSION_NAME,
)
from policy_sentry.shared.iam_data import (
    get_service_prefix_data,
    iam_definition,
)

logger = logging.getLogger(__name__)


@functools.lru_cache(maxsize=1024)
def get_all_service_prefixes() -> set[str]:
    """
    Gets all the AWS service prefixes from the actions table.

    If the action table does NOT have specific IAM actions (and therefore only supports * actions),
    then it will not be included in the response.

    Returns:
        List: A list of all AWS service prefixes present in the table.
    """
    results = set(iam_definition.keys())
    results.discard(POLICY_SENTRY_SCHEMA_VERSION_NAME)

    return results


@functools.lru_cache(maxsize=1024)
def get_all_actions(lowercase: bool = False) -> set[str]:
    """
    Gets a huge list of all IAM actions. This is used as part of the policyuniverse approach to minimizing
    IAM Policies to meet AWS-mandated character limits on policies.

    :param lowercase: Set to true to have the list of actions be in all lowercase strings.
    :return: A list of all actions present in the database.
    """
    all_service_prefixes = get_all_service_prefixes()

    all_actions: set[str] = set()

    for service_prefix in all_service_prefixes:
        service_prefix_data = get_service_prefix_data(service_prefix)
        if lowercase:
            action_names = service_prefix_data["privileges_lower_name"].keys()
        else:
            action_names = service_prefix_data["privileges_lower_name"].values()

        all_actions.update(f"{service_prefix}:{action_name}" for action_name in action_names)

    return all_actions


def get_service_authorization_url(service_prefix: str) -> str | None:
    """
    Gets the URL to the Actions, Resources, and Condition Keys page for a particular service.
    """
    result: str = iam_definition.get(service_prefix, {}).get("service_authorization_url")
    return result



import unittest
import json
import os
from policy_sentry.writing.sid_group import SidGroup

crud_with_override_template = os.path.abspath(
    os.path.join(
        os.path.dirname(__file__),
        os.path.pardir,
        os.path.pardir,
        "examples",
        "crud-with-override.yml",
    )
)


class SidGroupActionsTestCase(unittest.TestCase):
    def test_actions_test_case(self):
        cfg = {
            "mode": "actions",
            "name": "RoleNameWithCRUD",
            "actions": [
                "codestar-connections:UseConnection",
                "kms:CreateGrant",
                "kms:CreateCustomKeyStore",
                "ec2:AuthorizeSecurityGroupEgress",
                "ec2:AuthorizeSecurityGroupIngress",
            ],
        }
        sid_group = SidGroup()
        output = sid_group.process_template(cfg)
        print(json.dumps(output, indent=4))
        desired_output = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "CodestarconnectionsReadConnection",
                    "Effect": "Allow",
                    "Action": ["codestar-connections:UseConnection"],
                    "Resource": [
                        "arn:${Partition}:codestar-connections:${Region}:${Account}:connection/${ConnectionId}"
                    ],
                },
                {
                    "Sid": "KmsPermissionsmanagementKey",
                    "Effect": "Allow",
                    "Action": ["kms:CreateGrant"],
                    "Resource": ["arn:${Partition}:kms:${Region}:${Account}:key/${KeyId}"],
                },
                {
                    "Sid": "MultMultNone",
                    "Effect": "Allow",
                    "Action": ["cloudhsm:DescribeClusters", "kms:CreateCustomKeyStore"],
                    "Resource": ["*"],
                },
                {
                    "Sid": "Ec2WriteSecuritygroup",
                    "Effect": "Allow",
                    "Action": [
                        "ec2:AuthorizeSecurityGroupEgress",
                        "ec2:AuthorizeSecurityGroupIngress",
                    ],
                    "Resource": ["arn:${Partition}:ec2:${Region}:${Account}:security-group/${SecurityGroupId}"],
                },
            ],
        }
        expected_statement_ids = [
            "CodestarconnectionsReadConnection",
            "KmsPermissionsmanagementKey",
            "MultMultNone",
            "Ec2WriteSecuritygroup",
            "Ec2WriteSecuritygrouprule",
        ]
        for statement in output.get("Statement"):
            self.assertTrue(statement.get("Sid") in expected_statement_ids)



import unittest
import os
import json
from schema import Optional, Schema, And, Use, SchemaError
from policy_sentry.shared.iam_data import get_service_prefix_data
from policy_sentry.querying.actions import (
    get_actions_for_service,
    get_privilege_info,
    get_action_data,
    get_actions_that_support_wildcard_arns_only,
    get_actions_at_access_level_that_support_wildcard_arns_only,
    get_actions_with_access_level,
    get_actions_with_arn_type_and_access_level,
    remove_actions_not_matching_access_level,
    get_dependent_actions,
    remove_actions_that_are_not_wildcard_arn_only,
    get_actions_matching_condition_key,
    get_actions_matching_arn,
    get_actions_matching_arn_type,
    get_api_documentation_link_for_action,
    get_all_action_links,
    # get_actions_matching_condition_crud_and_arn
)
from policy_sentry.writing.validate import check


class QueryActionsTestCase(unittest.TestCase):
    def test_get_service_prefix_data(self):
        result = get_service_prefix_data("cloud9")
        desired_output_schema = Schema(
            {
                "service_name": "AWS Cloud9",
                "service_authorization_url": "https://docs.aws.amazon.com/service-authorization/latest/reference/list_awscloud9.html",
                "prefix": "cloud9",
                "privileges": dict,
                "privileges_lower_name": dict,
                "resources": dict,
                "resources_lower_name": dict,
                "conditions": dict,
            }
        )
        valid_output = check(desired_output_schema, result)
        # print(json.dumps(result, indent=4))
        self.assertTrue(valid_output)

    def test_get_actions_for_service(self):
        """querying.actions.get_actions_for_service"""
        expected_results = [
            "ram:AcceptResourceShareInvitation",
            "ram:AssociateResourceShare",
            "ram:AssociateResourceSharePermission",
            "ram:CreateResourceShare",
            "ram:DeleteResourceShare",
            "ram:DisassociateResourceShare",
            "ram:DisassociateResourceSharePermission",
            "ram:EnableSharingWithAwsOrganization",
            "ram:GetPermission",
            "ram:GetResourcePolicies",
            "ram:GetResourceShareAssociations",
            "ram:GetResourceShareInvitations",
            "ram:GetResourceShares",
            "ram:ListPendingInvitationResources",
            "ram:ListPermissions",
            "ram:ListPrincipals",
            "ram:ListResourceSharePermissions",
            "ram:ListResources",
            "ram:RejectResourceShareInvitation",
            "ram:TagResource",
            "ram:UntagResource",
            "ram:UpdateResourceShare",
        ]
        results = get_actions_for_service("ram")
        print(json.dumps(results, indent=4))
        self.maxDiff = None
        for expected_result in expected_results:
            self.assertTrue(expected_result in results)
        # self.assertListEqual(desired_output, output)

        # performance notes
        # old: 0.021s
        # this one: 0.005s

    def test_get_actions_for_invalid_service(self):
        """querying.actions.get_actions_for_service
        for invalid service
        """
        output = get_actions_for_service("invalid_service")
        self.assertListEqual([], output)

    def test_get_privilege_info(self):
        expected_results_file = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                "get_privilege_info_cloud9.json",
            )
        )
        # print(expected_results_file)
        with open(expected_results_file) as json_file:
            expected_results = json.load(json_file)
        results = get_privilege_info("cloud9", "CreateEnvironmentEC2")
        print(json.dumps(results, indent=4))
        # Future proofing the unit tests
        self.assertEqual(results["privilege"], expected_results["privilege"])
        self.assertEqual(results["access_level"], expected_results["access_level"])
        self.assertEqual(
            results["resource_types"][""]["resource_type"],
            expected_results["resource_types"][""]["resource_type"],
        )
        self.assertTrue("aws:RequestTag/${TagKey}" in results["resource_types"][""]["condition_keys"])
        self.assertTrue("ec2:DescribeSubnets" in results["resource_types"][""]["dependent_actions"])
        self.assertTrue("ec2:DescribeVpcs" in results["resource_types"][""]["dependent_actions"])
        self.assertTrue("iam:CreateServiceLinkedRole" in results["resource_types"][""]["dependent_actions"])
        self.assertTrue("environment" in results["service_resources"].keys())
        expected_service_conditions = [
            "aws:RequestTag/${TagKey}",
            "aws:ResourceTag/${TagKey}",
            "aws:TagKeys",
            "cloud9:EnvironmentId",
            "cloud9:EnvironmentName",
            "cloud9:InstanceType",
            "cloud9:OwnerArn",
            "cloud9:Permissions",
            "cloud9:SubnetId",
            "cloud9:UserArn",
        ]
        for expected_condition in expected_service_conditions:
            self.assertTrue(expected_condition in results["service_conditions"].keys())

    # def test_get_privilege_info_2(self):
    #     results = get_privilege_info("ram", "CreateResourceShare")
    #     # results = get_privilege_info("ram", "createresourceshare")
    #     print(json.dumps(results, indent=4))

    def test_get_action_data(self):
        """querying.actions.test_get_action_data"""
        desired_output = {
            "ram": [
                {
                    "action": "ram:TagResource",
                    "description": "Grants permission to tag the specified resource share or permission",
                    "access_level": "Tagging",
                    "api_documentation_link": "https://docs.aws.amazon.com/ram/latest/APIReference/API_TagResource.html",
                    "resource_arn_format": "arn:${Partition}:ram:${Region}:${Account}:permission/${ResourcePath}",
                    "condition_keys": [
                        "aws:ResourceTag/${TagKey}",
                        "ram:PermissionArn",
                        "ram:PermissionResourceType",
                        "aws:RequestTag/${TagKey}",
                        "aws:TagKeys",
                    ],
                    "dependent_actions": [],
                },
                {
                    "action": "ram:TagResource",
                    "description": "Grants permission to tag the specified resource share or permission",
                    "access_level": "Tagging",
                    "api_documentation_link": "https://docs.aws.amazon.com/ram/latest/APIReference/API_TagResource.html",
                    "resource_arn_format": "arn:${Partition}:ram:${Region}:${Account}:resource-share/${ResourcePath}",
                    "condition_keys": [
                        "aws:ResourceTag/${TagKey}",
                        "ram:AllowsExternalPrincipals",
                        "ram:ResourceShareName",
                        "aws:RequestTag/${TagKey}",
                        "aws:TagKeys",
                    ],
                    "dependent_actions": [],
                },
                {
                    "action": "ram:TagResource",
                    "description": "Grants permission to tag the specified resource share or permission",
                    "access_level": "Tagging",
                    "api_documentation_link": "https://docs.aws.amazon.com/ram/latest/APIReference/API_TagResource.html",
                    "resource_arn_format": "*",
                    "condition_keys": [
                        "aws:RequestTag/${TagKey}",
                        "aws:TagKeys",
                    ],
                    "dependent_actions": [],
                },
            ]
        }
        output = get_action_data("ram", "TagResource")
        print(json.dumps(output, indent=4))
        self.maxDiff = None
        self.assertDictEqual(desired_output, output)

    def test_get_action_data_with_glob(self):
        """Query action-table with glob."""
        desired_output = {
            "sns": [
                {
                    "action": "sns:ListSubscriptions",
                    "description": "Grants permission to return a list of the requester's subscriptions",
                    "access_level": "List",
                    "api_documentation_link": "https://docs.aws.amazon.com/sns/latest/api/API_ListSubscriptions.html",
                    "resource_arn_format": "*",
                    "condition_keys": [],
                    "dependent_actions": [],
                },
                {
                    "action": "sns:ListSubscriptionsByTopic",
                    "description": "Grants permission to return a list of the subscriptions to a specific topic",
                    "access_level": "List",
                    "api_documentation_link": "https://docs.aws.amazon.com/sns/latest/api/API_ListSubscriptionsByTopic.html",
                    "resource_arn_format": "arn:${Partition}:sns:${Region}:${Account}:${TopicName}",
                    "condition_keys": ["aws:ResourceTag/${TagKey}"],
                    "dependent_actions": [],
                },
            ]
        }
        results = get_action_data("sns", "ListSubscriptions*")
        self.assertDictEqual(desired_output, results)

    def test_get_actions_that_support_wildcard_arns_only(self):
        """querying.actions.get_actions_that_support_wildcard_arns_only"""
        # Variant 1: Secrets manager
        expected_results = [
            "secretsmanager:BatchGetSecretValue",
            "secretsmanager:GetRandomPassword",
            "secretsmanager:ListSecrets",
        ]
        results = get_actions_that_support_wildcard_arns_only("secretsmanager")
        self.maxDiff = None
        self.assertCountEqual(expected_results, results)

        # Variant 2: ECR
        results = get_actions_that_support_wildcard_arns_only("ecr")
        expected_results = [
            "ecr:DeleteRegistryPolicy",
            "ecr:DescribeRegistry",
            "ecr:GetAuthorizationToken",
            "ecr:GetRegistryPolicy",
            "ecr:PutRegistryPolicy",
            "ecr:PutReplicationConfiguration",
        ]
        # print(json.dumps(results, indent=4))
        for item in expected_results:
            self.assertTrue(item in results)

        # Variant 3: All actions
        output = get_actions_that_support_wildcard_arns_only("all")
        # print(len(output))
        self.assertTrue(len(output) > 3000)

    def test_get_actions_at_access_level_that_support_wildcard_arns_only(self):
        """querying.actions.get_actions_at_access_level_that_support_wildcard_arns_only"""
        read_output = get_actions_at_access_level_that_support_wildcard_arns_only("secretsmanager", "Read")
        list_output = get_actions_at_access_level_that_support_wildcard_arns_only("secretsmanager", "List")
        write_output = get_actions_at_access_level_that_support_wildcard_arns_only("secretsmanager", "Write")
        tagging_output = get_actions_at_access_level_that_support_wildcard_arns_only("secretsmanager", "Tagging")
        permissions_output = get_actions_at_access_level_that_support_wildcard_arns_only("s3", "Permissions management")
        # print(json.dumps(read_output, indent=4))
        # print(json.dumps(list_output, indent=4))
        # print(json.dumps(write_output, indent=4))
        # print(json.dumps(tagging_output, indent=4))
        # print(json.dumps(permissions_output, indent=4))
        self.assertListEqual(read_output, ["secretsmanager:GetRandomPassword"])
        self.assertListEqual(
            list_output,
            ["secretsmanager:BatchGetSecretValue", "secretsmanager:ListSecrets"],
        )
        self.assertListEqual(write_output, [])
        self.assertListEqual(tagging_output, [])
        for item in ["s3:PutAccountPublicAccessBlock"]:
            self.assertTrue(item in permissions_output)
        all_permissions_output = get_actions_at_access_level_that_support_wildcard_arns_only(
            "all", "Permissions management"
        )
        all_write_output = get_actions_at_access_level_that_support_wildcard_arns_only("all", "Write")

        print(len(all_permissions_output) + len(all_write_output))
        # print(len(all_write_output))
        # print(json.dumps(all_write_output, indent=4))
        # print(json.dumps(all_permissions_output, indent=4))

    def test_get_actions_with_access_level(self):
        """querying.actions.get_actions_with_access_level"""
        desired_output = ["workspaces:CreateTags", "workspaces:DeleteTags"]
        output = get_actions_with_access_level("workspaces", "Tagging")
        self.maxDiff = None
        self.assertListEqual(desired_output, output)
        # output = get_actions_with_access_level(
        #     "all", "Tagging"
        # )
        # print(output)

    def test_get_actions_with_arn_type_and_access_level_case_1(self):
        """querying.actions.get_actions_with_arn_type_and_access_level"""
        desired_output = [
            "s3:DeleteBucketPolicy",
            "s3:PutBucketAcl",
            "s3:PutBucketOwnershipControls",
            "s3:PutBucketPolicy",
            "s3:PutBucketPublicAccessBlock",
        ]
        output = get_actions_with_arn_type_and_access_level(
            # "ram", "resource-share", "Write"
            "s3",
            "bucket",
            "Permissions management",
        )
        self.maxDiff = None
        self.assertListEqual(desired_output, output)

    def test_get_actions_with_arn_type_and_access_level_case_2(self):
        """querying.actions.get_actions_with_arn_type_and_access_level with arn type"""
        desired_output = [
            "ssm:DeleteParameter",
            "ssm:DeleteParameters",
            "ssm:LabelParameterVersion",
            "ssm:PutParameter",
        ]
        output = get_actions_with_arn_type_and_access_level("ssm", "parameter", "Write")
        for item in desired_output:
            self.assertTrue(item in output)

    def test_get_actions_with_arn_type_and_access_level_case_3(self):
        """querying.actions.get_actions_with_arn_type_and_access_level with arn type"""
        desired_output = [
            "s3:PutAccountPublicAccessBlock",
            "s3:PutAccessPointPublicAccessBlock",
        ]
        output = get_actions_with_arn_type_and_access_level(
            # "ram", "resource-share", "Write"
            "s3",
            "*",
            "Permissions management",
        )
        print(output)
        for item in desired_output:
            self.assertTrue(item in output)
        # self.assertListEqual(desired_output, output)

    def test_get_actions_with_arn_type_and_access_level_case_4(self):
        """querying.actions.get_actions_with_arn_type_and_access_level with arn type"""
        desired_output = [
            "secretsmanager:BatchGetSecretValue",
            "secretsmanager:ListSecrets",
        ]
        output = get_actions_with_arn_type_and_access_level("secretsmanager", "*", "List")
        self.assertCountEqual(desired_output, output)

    def test_get_actions_with_arn_type_and_access_level_case_5(self):
        """querying.actions.get_actions_with_arn_type_and_access_level with arn type"""

        output = get_actions_with_arn_type_and_access_level("s3", "object", "List")
        self.assertTrue("s3:ListMultipartUploadParts" in output)

    def test_get_actions_with_arn_type_and_access_level_servicecatalog(self):
        # querying with `catalog` as a service prefix should yield `servicecatalog` as an action prefix
        # https://github.com/salesforce/policy_sentry/issues/563

        output = get_actions_with_arn_type_and_access_level("catalog", "Portfolio", "List")
        self.assertIn("servicecatalog:ListPortfolioAccess", output)

    def test_get_actions_matching_arn_type_case_1(self):
        """querying.actions.get_actions_matching_arn_type"""
        expected_results = [
            "ecr:DeleteRegistryPolicy",
            "ecr:DescribeRegistry",
            "ecr:GetAuthorizationToken",
            "ecr:GetRegistryPolicy",
            "ecr:PutRegistryPolicy",
            "ecr:PutReplicationConfiguration",
        ]
        results = get_actions_matching_arn_type("ecr", "*")
        print(json.dumps(results, indent=4))
        for item in expected_results:
            self.assertTrue(item in results)
        # self.assertEqual(output, ["ecr:GetAuthorizationToken"])

    def test_get_actions_matching_arn_type_case_2(self):
        """querying.actions.get_actions_matching_arn_type"""
        output = get_actions_matching_arn_type("all", "object")
        self.assertTrue("s3:AbortMultipartUpload" in output)

    def test_get_actions_matching_arn_type_case_3(self):
        """querying.actions.get_actions_matching_arn_type"""
        output = get_actions_matching_arn_type("rds", "object")
        self.assertTrue(len(output) == 0)

    def test_get_actions_matching_arn_type_case_4(self):
        """querying.actions.get_actions_matching_arn_type"""

        desired_output = [
            "codestar:CreateUserProfile",
            "codestar:DeleteUserProfile",
            "codestar:UpdateUserProfile",
        ]

        output = get_actions_matching_arn_type("codestar", "user")

        self.assertTrue(output, desired_output)

    def test_get_actions_matching_condition_key(self):
        """querying.actions.get_actions_matching_condition_key"""

        results = get_actions_matching_condition_key("ses", "ses:FeedbackAddress")
        expected_results = [
            # 'ses:SendBulkTemplatedEmail',
            # 'ses:SendCustomVerificationEmail',
            "ses:SendEmail",
            # 'ses:SendRawEmail',
            # 'ses:SendTemplatedEmail'
        ]
        # print(output)
        self.maxDiff = None
        print(results)
        for expected_result in expected_results:
            self.assertTrue(expected_result in results)
        # self.assertListEqual(results, desired_results)

    # def test_get_actions_matching_condition_crud_and_arn(self):
    #     """querying.actions.get_actions_matching_condition_crud_and_arn"""
    #     results = get_actions_matching_condition_crud_and_arn(
    #         "elasticbeanstalk:InApplication",
    #         "List",
    #         "arn:${Partition}:elasticbeanstalk:${Region}:${Account}:environment/${ApplicationName}/${EnvironmentName}",
    #     )
    #     desired_results = [
    #         "elasticbeanstalk:DescribeEnvironments",
    #     ]
    #     print(results)
    #     self.assertListEqual(results, desired_results)
    #
    # def test_get_actions_matching_condition_crud_and_wildcard_arn(self):
    #     """querying.actions.get_actions_matching_condition_crud_and_wildcard_arn"""
    #     desired_results = [
    #         "swf:PollForActivityTask",
    #         "swf:PollForDecisionTask",
    #         "swf:RespondActivityTaskCompleted",
    #         "swf:StartWorkflowExecution",
    #     ]
    #     results = get_actions_matching_condition_crud_and_arn(
    #         "swf:taskList.name", "Write", "*"
    #     )
    #     print(results)
    #     self.assertListEqual(desired_results, results)
    #
    #     # This one leverages a condition key that is partway through a string in the database
    #     # - luckily, SQLAlchemy's ilike function allows us to find it anyway because it's a substring
    #     # kms:CallerAccount,kms:EncryptionAlgorithm,kms:EncryptionContextKeys,kms:ViaService
    #     desired_results = [
    #         "kms:Decrypt",
    #         "kms:Encrypt",
    #         "kms:GenerateDataKey",
    #         "kms:GenerateDataKeyPair",
    #         "kms:GenerateDataKeyPairWithoutPlaintext",
    #         "kms:GenerateDataKeyWithoutPlaintext",
    #         "kms:ReEncryptFrom",
    #         "kms:ReEncryptTo",
    #     ]
    #     print(results)
    #     results = get_actions_matching_condition_crud_and_arn(
    #         "kms:EncryptionAlgorithm", "Write", "*"
    #     )
    #     self.assertListEqual(desired_results, results)

    def test_remove_actions_not_matching_access_level(self):
        # TODO: This method normalized the access level unnecessarily. Make sure to change that in the final iteration
        """querying.actions.remove_actions_not_matching_access_level"""
        actions_list = [
            "ecr:batchgetimage",  # read
            "ecr:createrepository",  # write
            "ecr:describerepositories",  # list
            "ecr:tagresource",  # tagging
            "ecr:setrepositorypolicy",  # permissions management
        ]
        self.maxDiff = None
        # Read
        result = remove_actions_not_matching_access_level(actions_list, "Read")
        expected = ["ecr:BatchGetImage"]
        self.assertListEqual(result, ["ecr:BatchGetImage", "ecr:DescribeRepositories"])
        # Write
        result = remove_actions_not_matching_access_level(actions_list, "Write")

        self.assertListEqual(result, ["ecr:CreateRepository"])
        # List
        result = remove_actions_not_matching_access_level(actions_list, "List")
        # self.assertListEqual(result, ["ecr:DescribeRepositories"])
        # DescribeRepositories is no longer considered a "list" action.
        self.assertListEqual(result, [])

        # Tagging
        result = remove_actions_not_matching_access_level(actions_list, "Tagging")
        self.assertListEqual(result, ["ecr:TagResource"])
        # Permissions management
        result = remove_actions_not_matching_access_level(actions_list, "Permissions management")
        self.assertListEqual(result, ["ecr:SetRepositoryPolicy"])

        bad_actions_list = [
            "codecommit:CreatePullRequest",
            "codecommit:CreatePullRequestApprovalRule",
            "codecommit:CreateRepository",
            "codecommit:CreateUnreferencedMergeCommit",
            "codecommit:DeleteBranch",
            "codecommit:DeleteFile",
        ]

    def test_get_dependent_actions(self):
        """querying.actions.get_dependent_actions"""
        dependent_actions_single = ["ec2:associateiaminstanceprofile"]
        dependent_actions_double = ["shield:associatedrtlogbucket"]
        dependent_actions_several = ["chime:getcdrbucket"]
        self.assertEqual(
            get_dependent_actions(dependent_actions_single),
            ["iam:PassRole"],
        )
        self.assertCountEqual(
            get_dependent_actions(dependent_actions_double),
            ["s3:GetBucketPolicy", "s3:PutBucketPolicy"],
        )
        self.assertCountEqual(
            get_dependent_actions(dependent_actions_several),
            [
                "s3:GetBucketAcl",
                "s3:GetBucketLocation",
                "s3:GetBucketLogging",
                "s3:GetBucketVersioning",
                "s3:GetBucketWebsite",
            ],
        )

    def test_remove_actions_that_are_not_wildcard_arn_only(self):
        """querying.actions.remove_actions_that_are_not_wildcard_arn_only"""
        provided_actions_list = [
            # 3 wildcard only actions
            "secretsmanager:getrandompassword",
            "secretsmanager:listsecrets",
            # These ones are wildcard OR "secret"
            "secretsmanager:createsecret",
            "secretsmanager:putsecretvalue",
        ]
        desired_output = [
            # 2 wildcard only actions
            "secretsmanager:GetRandomPassword",
            "secretsmanager:ListSecrets",
        ]
        output = remove_actions_that_are_not_wildcard_arn_only(provided_actions_list)
        self.maxDiff = None
        self.assertCountEqual(desired_output, output)

    def test_weird_lowercase_uppercase(self):
        """test_weird_lowercase_uppercase: Same as test_remove_actions_that_are_not_wildcard_arn_only, but with wEiRd cases"""
        provided_actions_list = [
            # 2 wildcard only actions
            "secretsmanager:gEtRaNdOmPasSwOrD",
            "secretsmanager:LIstsEcretS",
            # This one is wildcard OR "secret"
            "secretsmanager:cReAtEsEcReT",
            "secretsmanager:pUtSeCrEtVaLuE",
        ]
        desired_output = [
            # 2 wildcard only actions
            "secretsmanager:GetRandomPassword",
            "secretsmanager:ListSecrets",
        ]
        output = remove_actions_that_are_not_wildcard_arn_only(provided_actions_list)
        print(json.dumps(output))
        self.maxDiff = None
        self.assertCountEqual(desired_output, output)

    def test_get_actions_matching_arn(self):
        """querying.actions.get_actions_matching_arn"""
        arn = "arn:aws:cloud9:us-east-1:account-id:environment:123456"
        results = get_actions_matching_arn(arn)
        # print(json.dumps(results, indent=4))
        # Don't want to keep an updated list of actions in these tests,
        # so let's just test the lengths and look for some contents that should or should not be in there.
        self.assertTrue(len(results) > 10)
        self.assertTrue("cloud9:ListEnvironments" not in results)
        self.assertTrue("cloud9:DeleteEnvironment" in results)

    def test_gh_226_elasticloadbalancing_v1_and_v2(self):
        """Test that elasticloadbalancing combines v1 and v2"""
        results = get_actions_for_service("elasticloadbalancing")
        # print(json.dumps(results, indent=4))
        lb_v1_only_action = "elasticloadbalancing:CreateTargetGroup"
        lb_v2_only_action = "elasticloadbalancing:SetSecurityGroups"
        self.assertTrue(lb_v1_only_action in results)
        self.assertTrue(lb_v2_only_action in results)

    def test_get_api_documentation_link_for_action(self):
        """querying.actions.get_api_documentation_link_for_action"""
        service_prefix = "cloud9"
        action_name = "CreateEnvironmentEC2"
        result = get_api_documentation_link_for_action(service_prefix, action_name)
        print(result)
        # Link should be: https://docs.aws.amazon.com/cloud9/latest/APIReference/API_CreateEnvironmentEC2.html
        # We will just check the https and subdomain.domain in case they change the format in the future.
        self.assertTrue("https://docs.aws.amazon.com" in result)

    def test_get_all_links(self):
        """querying.actions.get_all_action_links"""
        results = get_all_action_links()
        self.assertTrue(len(results.keys()) > 8000)

    # def test_research_access_levels(self):
    #     """querying.actions.get_actions_with_access_level"""
    #     service_prefix = "sns"
    #     access_level = "Write"
    #     # access_level = "Permissions management"
    #     results = get_actions_with_access_level(service_prefix, access_level)
    #     print(json.dumps(results, indent=4))



from policy_sentry.analysis.analyze import (
    determine_actions_to_expand,
    analyze_by_access_level,
    analyze_statement_by_access_level,
)
import unittest
import json


class AnalysisExpandWildcardActionsTestCase(unittest.TestCase):
    def test_a_determine_actions_to_expand_not_upper_camelcase(self):
        """test_determine_actions_to_expand_not_upper_camelcase: The nOtCaMeLcAsE version of the same test"""
        action_list = ["ecr:pUt*"]
        self.maxDiff = None
        expected_results = [
            "ecr:PutImage",
            "ecr:PutImageScanningConfiguration",
            "ecr:PutImageTagMutability",
            "ecr:PutLifecyclePolicy",
        ]
        result = determine_actions_to_expand(action_list)
        print(result)
        self.maxDiff = None
        results = determine_actions_to_expand(action_list)
        for expected_result in expected_results:
            self.assertTrue(expected_result in results)

    def test_determine_actions_to_expand(self):
        """
        test_determine_actions_to_expand: provide expanded list of actions, like ecr:*
        :return:
        """
        action_list = ["ecr:*"]
        self.maxDiff = None
        expected_results = [
            "ecr:BatchCheckLayerAvailability",
            "ecr:BatchDeleteImage",
            "ecr:BatchGetImage",
            "ecr:CompleteLayerUpload",
            "ecr:CreateRepository",
            "ecr:DeleteLifecyclePolicy",
            "ecr:DeleteRepository",
            "ecr:DeleteRepositoryPolicy",
            "ecr:DescribeImageScanFindings",
            "ecr:DescribeImages",
            "ecr:DescribeRepositories",
            "ecr:GetAuthorizationToken",
            "ecr:GetDownloadUrlForLayer",
            "ecr:GetLifecyclePolicy",
            "ecr:GetLifecyclePolicyPreview",
            "ecr:GetRepositoryPolicy",
            "ecr:InitiateLayerUpload",
            "ecr:ListImages",
            "ecr:ListTagsForResource",
            "ecr:PutImage",
            "ecr:PutImageScanningConfiguration",
            "ecr:PutImageTagMutability",
            "ecr:PutLifecyclePolicy",
            "ecr:SetRepositoryPolicy",
            "ecr:StartImageScan",
            "ecr:StartLifecyclePolicyPreview",
            "ecr:TagResource",
            "ecr:UntagResource",
            "ecr:UploadLayerPart",
        ]
        # print(determine_actions_to_expand(action_list))
        self.maxDiff = None
        results = determine_actions_to_expand(action_list)
        for expected_result in expected_results:
            self.assertTrue(expected_result in results)

    def test_analyze_by_access_level(self):
        """test_analyze_by_access_level: Test out calling this as a library"""
        permissions_management_policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": [
                        # This one is Permissions management
                        "ecr:SetRepositoryPolicy",
                        "secretsmanager:DeleteResourcePolicy",
                        # These ones are not permissions management
                        "ecr:getrepositorypolicy",
                        "ecr:describerepositories",
                        "ecr:listimages",
                        "ecr:DescribeImages",
                    ],
                    "Resource": "*",
                },
                {
                    "Sid": "AllowManageOwnAccessKeys",
                    "Effect": "Allow",
                    "Action": [
                        # These ones are permissions management
                        "iam:CreateAccessKey",
                        "iam:DeleteAccessKey",
                        "iam:UpdateAccessKey",
                        # This one is not
                        "iam:ListAccessKeys",
                    ],
                    "Resource": "arn:aws:iam::*:user/${aws:username}",
                },
            ],
        }
        permissions_management_actions = analyze_by_access_level(
            permissions_management_policy, "Permissions management"
        )
        # print(permissions_management_actions)
        desired_actions_list = [
            "ecr:SetRepositoryPolicy",
            "iam:CreateAccessKey",
            "iam:DeleteAccessKey",
            "iam:UpdateAccessKey",
            "secretsmanager:DeleteResourcePolicy",
        ]
        self.maxDiff = None
        self.assertListEqual(permissions_management_actions, desired_actions_list)

    def test_analyze_statement_by_access_level(self):
        """test_analyze_statement_by_access_level: Test out calling this as a library"""
        permissions_management_statement = {
            "Effect": "Allow",
            "Action": [
                # This one is Permissions management
                "ecr:SetRepositoryPolicy",
                "secretsmanager:DeleteResourcePolicy",
                # These ones are not permissions management
                "ecr:GetRepositoryPolicy",
                "ecr:DescribeRepositories",
                "ecr:ListImages",
                "ecr:DescribeImages",
            ],
            "Resource": "*",
        }
        result = analyze_statement_by_access_level(permissions_management_statement, "Permissions management")
        # print(permissions_management_actions)
        desired_result = [
            "ecr:SetRepositoryPolicy",
            "secretsmanager:DeleteResourcePolicy",
        ]
        self.maxDiff = None
        self.assertListEqual(result, desired_result)

    def test_gh_162(self):
        """test_gh_162: Addressing the concern in the Github issue
        https://github.com/salesforce/policy_sentry/issues/162"""
        permissions_management_policy = {
            "Statement": [
                {
                    "Action": ["s3:GetObject*", "s3:PutObject*"],
                    "Effect": "Allow",
                    "Resource": ["*"],
                }
            ],
            "Version": "2012-10-17",
        }
        print("********* READ ***********")
        results = analyze_by_access_level(permissions_management_policy, "Read")
        print(json.dumps(results, indent=4))
        # Rather than maintaining a large list as AWS keeps adding new actions,
        # just verify that an expanded action exists in the list
        self.assertTrue("s3:GetObjectAcl" in results)
        print("********* LIST ***********")
        results = analyze_by_access_level(permissions_management_policy, "List")
        print(json.dumps(results, indent=4))
        self.assertListEqual(results, [])
        print("********* WRITE ***********")
        results = analyze_by_access_level(permissions_management_policy, "Write")
        print(json.dumps(results, indent=4))
        self.assertTrue("s3:PutObjectLegalHold" in results)
        print("********* TAGGING ***********")
        results = analyze_by_access_level(permissions_management_policy, "Tagging")
        print(json.dumps(results, indent=4))
        self.assertTrue("s3:PutObjectTagging" in results)
        print("********* PERMISSIONS-MANAGEMENT ***********")
        results = analyze_by_access_level(permissions_management_policy, "Permissions management")
        print(json.dumps(results, indent=4))
        self.assertTrue("s3:PutObjectAcl" in results)



"""
Create the Policy Sentry config folder (~/.policy_sentry/) and the contents within
Create the SQLite datastore and fill it with the tables scraped from the AWS Docs
"""

from __future__ import annotations

import logging
import shutil
from pathlib import Path

import click

from policy_sentry import set_stream_logger
from policy_sentry.querying.all import get_all_service_prefixes
from policy_sentry.shared.constants import (
    BUNDLED_DATA_DIRECTORY,
    BUNDLED_DATASTORE_FILE_PATH,
    CONFIG_DIRECTORY,
    DATASTORE_FILE_PATH,
    LOCAL_ACCESS_OVERRIDES_FILE,
    LOCAL_DATASTORE_FILE_PATH,
    LOCAL_HTML_DIRECTORY_PATH,
)

logger = logging.getLogger(__name__)


@click.command(
    name="initialize",
    short_help="Create a local datastore to store AWS IAM information.",
)
@click.option(
    "--access-level-overrides-file",
    type=str,
    required=False,
    help="Path to access level overrides file, used to override the Access Levels per action provided by AWS docs",
)
@click.option(
    "--fetch",
    is_flag=True,
    required=False,
    default=False,
    help="Specify this flag to fetch the HTML Docs directly from the AWS website. This will be helpful if the docs "
    "in the Git repository are behind the live docs and you need to use the latest version of the docs right "
    "now.",
)
@click.option(
    "--build",
    is_flag=True,
    required=False,
    default=False,
    help="Build the IAM data file from the HTML files rather than copying the data file from "
    "the python package. Defaults to false",
)
@click.option(
    "--verbose",
    "-v",
    type=click.Choice(["critical", "error", "warning", "info", "debug"], case_sensitive=False),
)
def initialize_command(
    access_level_overrides_file: str | None,
    fetch: bool,
    build: bool,
    verbose: str | None,
) -> None:
    """
    CLI command for initializing the local data file
    """
    if verbose:
        log_level = getattr(logging, verbose.upper())
        set_stream_logger(level=log_level)

    initialize(access_level_overrides_file, fetch, build)


def initialize(
    access_level_overrides_file: str | None = None,
    fetch: bool = False,
    build: bool = False,
) -> None:
    """
    Initialize the local data file to store AWS IAM information, which can be used to generate IAM policies, and for
    querying the database.
    """

    # importing 'awsdocs' is quite pricey, when it is actually only used for initialize the IAM DB
    from policy_sentry.shared.awsdocs import (
        create_database,
        update_html_docs_directory,
    )

    if not access_level_overrides_file:
        overrides_file = LOCAL_ACCESS_OVERRIDES_FILE
    else:
        overrides_file = Path(access_level_overrides_file)
    # Create the config directory
    database_path = create_policy_sentry_config_directory()

    # Create local html docs folder, if it doesn't exist
    LOCAL_HTML_DIRECTORY_PATH.mkdir(parents=True, exist_ok=True)

    # Create overrides file, which allows us to override the Access Levels
    # provided by AWS documentation
    file_list = [f for f in BUNDLED_DATA_DIRECTORY.iterdir() if (BUNDLED_DATA_DIRECTORY / f).is_file()]

    for file in file_list:
        if file.suffix == ".yml":
            shutil.copy(BUNDLED_DATA_DIRECTORY / file, CONFIG_DIRECTORY)
            logger.debug("copying overrides file %s to %s", file, CONFIG_DIRECTORY)
    print(f"Database will be stored here: {database_path}")

    if not build and not fetch:
        # copy from the bundled database location to the destination path
        shutil.copy(BUNDLED_DATASTORE_FILE_PATH, database_path)

    # --fetch: wget the AWS IAM Actions, Resources and Condition Keys pages and store them locally.
    if fetch:
        # `wget` the html docs to the local directory
        update_html_docs_directory(LOCAL_HTML_DIRECTORY_PATH)
    elif not next(LOCAL_HTML_DIRECTORY_PATH.glob("*.html"), None):
        print("No HTML docs found, fetching from AWS!")
        update_html_docs_directory(LOCAL_HTML_DIRECTORY_PATH)

    # --build
    if build or access_level_overrides_file or fetch:
        create_database(CONFIG_DIRECTORY, overrides_file)
        print("Created the database!")

    # Query the database for all the services that are now in the database.
    all_aws_service_prefixes = get_all_service_prefixes()
    print("Initialization complete!")
    print(f"Total AWS services in the IAM database: {len(all_aws_service_prefixes)}")
    logger.debug("\nService prefixes:")
    logger.debug(", ".join(all_aws_service_prefixes))


def create_policy_sentry_config_directory() -> Path:
    """
    Creates a config directory at $HOME/.policy_sentry/
    :return: the path of the database file
    """
    print("Creating the database...")
    logger.debug(f"We will store the new database here: {DATASTORE_FILE_PATH}")
    # If the database file already exists, remove it
    if LOCAL_DATASTORE_FILE_PATH.exists():
        logger.debug(f"The database at {DATASTORE_FILE_PATH} already exists. Removing and replacing it.")
        LOCAL_DATASTORE_FILE_PATH.unlink()
    else:
        CONFIG_DIRECTORY.mkdir(exist_ok=True)
    return LOCAL_DATASTORE_FILE_PATH



import unittest
from policy_sentry.analysis.expand import get_expanded_policy


class PolicyExpansionTestCase(unittest.TestCase):
    def test_policy_expansion(self):
        """command.expand_policy.get_expanded_policy: Test the expansion of the cloud9 service"""
        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "TestSID",
                    "Effect": "Allow",
                    "Action": ["cloud9:*"],
                    "Resource": "*",
                }
            ],
        }
        output = get_expanded_policy(policy)
        # print(json.dumps(output, indent=4))
        desired_output = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "TestSID",
                    "Effect": "Allow",
                    "Action": [
                        "cloud9:CreateEnvironmentEC2",
                        "cloud9:CreateEnvironmentMembership",
                        "cloud9:DeleteEnvironment",
                        "cloud9:DeleteEnvironmentMembership",
                        "cloud9:DescribeEnvironmentMemberships",
                        "cloud9:DescribeEnvironmentStatus",
                        "cloud9:DescribeEnvironments",
                        "cloud9:GetUserSettings",
                        "cloud9:ListEnvironments",
                        "cloud9:ListTagsForResource",
                        "cloud9:TagResource",
                        "cloud9:UntagResource",
                        "cloud9:UpdateEnvironment",
                        "cloud9:UpdateEnvironmentMembership",
                        "cloud9:UpdateUserSettings",
                    ],
                    "Resource": "*",
                }
            ],
        }
        expected_statement_ids = ["TestSID"]
        for statement in output.get("Statement"):
            self.assertTrue(statement.get("Sid") in expected_statement_ids)
        expected_actions = [
            "cloud9:CreateEnvironmentEC2",
            "cloud9:CreateEnvironmentMembership",
            "cloud9:DeleteEnvironment",
            "cloud9:DeleteEnvironmentMembership",
            "cloud9:DescribeEnvironmentMemberships",
            "cloud9:DescribeEnvironmentStatus",
            "cloud9:DescribeEnvironments",
            "cloud9:GetUserSettings",
            "cloud9:ListEnvironments",
            "cloud9:ListTagsForResource",
            "cloud9:TagResource",
            "cloud9:UntagResource",
            "cloud9:UpdateEnvironment",
            "cloud9:UpdateEnvironmentMembership",
            "cloud9:UpdateUserSettings",
        ]
        for action in expected_actions:
            self.assertTrue(action in output["Statement"][0]["Action"])



import json
import unittest
from click.testing import CliRunner
from policy_sentry.command.query import query


class QueryClickUnitTests(unittest.TestCase):
    def setUp(self):
        self.runner = CliRunner()

    def test_query_action_table_basic_with_click(self):
        """command.query.query_action_table: should return exit code 0"""
        result = self.runner.invoke(query, ["action-table", "--service", "ram"])
        self.assertTrue(result.exit_code == 0)

    def test_click_query_action_table(self):
        """command.query.query_action_table: click testing"""
        cases = [
            [
                "action-table",
                "--service",
                "all",
                "--access-level",
                "permissions-management",
            ],
            ["action-table", "--service", "all", "--resource-type", "*"],
            ["action-table", "--service", "all"],
            [
                "action-table",
                "--service",
                "ram",
                "--access-level",
                "permissions-management",
            ],
            [
                "action-table",
                "--service",
                "ram",
                "--access-level",
                "permissions-management",
                "--resource-type",
                "resource-share",
            ],
            ["action-table", "--service", "ses", "--condition", "ses:FeedbackAddress"],
            ["action-table", "--service", "ssm", "--resource-type", "*"],
            ["action-table", "--service", "ram", "--name", "tagresource"],
            ["action-table", "--service", "ssm", "--resource-type", "parameter"],
            ["action-table", "--service", "ram"],
        ]
        for case in cases:
            result = self.runner.invoke(query, case)
            print(result.output)
            self.assertTrue(result.exit_code == 0)

    def test_click_query_arn_table(self):
        """command.query.query_arn_table: click testing"""
        cases = [
            ["arn-table", "--service", "ssm"],
            ["arn-table", "--service", "cloud9", "--list-arn-types"],
            ["arn-table", "--service", "cloud9", "--name", "environment"],
        ]
        for case in cases:
            result = self.runner.invoke(query, case)
            print(result.output)
            self.assertTrue(result.exit_code == 0)

    def test_click_query_condition_table(self):
        """command.query.query_condition_table: click testing"""
        cases = [
            ["condition-table", "--service", "cloud9"],
            ["condition-table", "--service", "cloud9", "--name", "cloud9:Permissions"],
        ]
        for case in cases:
            result = self.runner.invoke(query, case)
            print(result.output)
            self.assertTrue(result.exit_code == 0)



import unittest
import json
import os
from policy_sentry.command.write_policy import write_policy_with_template
from policy_sentry.util.file import read_yaml_file
from policy_sentry.util.policy_files import (
    get_sid_names_from_policy,
    get_statement_from_policy_using_sid,
)


# class WritePolicyPreventWildcardEscalation(unittest.TestCase):
#     def test_wildcard_when_not_necessary(self):
#         """test_wildcard_when_not_necessary: Attempts bypass of CRUD mode wildcard-only"""
#         cfg = {
#             "mode": "crud",
#             "name": "RoleNameWithCRUD",
#             "permissions-management": ["arn:aws:s3:::example-org-s3-access-logs"],
#             "wildcard-only": {
#                 "single-actions": [
#                     # The first three are legitimately wildcard only.
#                     # Verify with `policy_sentry query action-table --service secretsmanager --wildcard-only`
#                     "ram:EnableSharingWithAwsOrganization",
#                     "ram:GetResourcePolicies",
#                     "secretsmanager:CreateSecret",
#                     # This last one can be "secret" ARN type OR wildcard. We want to prevent people from
#                     # bypassing this mechanism, while allowing them to explicitly
#                     # request specific privs that require wildcard mode. This next value -
#                     # secretsmanager:putsecretvalue - is an example of someone trying to beat the tool.
#                     "secretsmanager:PutSecretValue",
#                 ],
#             }
#         }
#         output = write_policy_with_template(cfg)
#         # print(json.dumps(output, indent=4))
#         desired_output = {
#             "Version": "2012-10-17",
#             "Statement": [
#                 {
#                     "Sid": "MultMultNone",
#                     "Effect": "Allow",
#                     "Action": [
#                         "ram:EnableSharingWithAwsOrganization",
#                         "ram:GetResourcePolicies",
#                     ],
#                     "Resource": ["*"],
#                 },
#                 {
#                     "Sid": "S3PermissionsmanagementBucket",
#                     "Effect": "Allow",
#                     "Action": [
#                         "s3:DeleteBucketPolicy",
#                         "s3:PutBucketAcl",
#                         "s3:PutBucketPolicy",
#                         "s3:PutBucketPublicAccessBlock",
#                     ],
#                     "Resource": ["arn:aws:s3:::example-org-s3-access-logs"],
#                 },
#             ],
#         }
#         self.maxDiff = None
#         print(output)
#         self.assertDictEqual(desired_output, output)


# class TaggingTestCase(unittest.TestCase):
#     def test_write_tagging_only_policy(self):
#         """test_write_tagging_only_policy: We'd never write a policy like this IRL but doing this as a quality check against how it handles the database"""
#         policy_file_path = abspath(
#             join(
#                 dirname(__file__), pardir + "/" + pardir + "/examples/yml/tagging.yml",
#             )
#         )
#         cfg = read_yaml_file(policy_file_path)
#
#         policy = write_policy_with_template(cfg)
#         print(policy)


class WildcardOnlyServiceLevelTestCase(unittest.TestCase):
    def test_add_wildcard_only_actions_matching_services_and_access_level(self):
        """test_add_wildcard_only_actions_matching_services_and_access_level: We'd never write a policy like this
        IRL but doing this as a quality check against how it handles the database"""
        policy_file_path = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                os.path.pardir,
                os.path.pardir,
                "examples",
                "yml",
                "crud-with-wildcard-service-level.yml",
            )
        )
        cfg = read_yaml_file(policy_file_path)

        results = write_policy_with_template(cfg)

        # The Policy *should* look like this.

        # desired_output = {
        #     "Version": "2012-10-17",
        #     "Statement": [
        #         {
        #             "Sid": "MultMultNone",
        #             "Effect": "Allow",
        #             "Action": [
        #                 "ram:EnableSharingWithAwsOrganization",
        #                 "ram:GetResourcePolicies",
        #                 "ecr:GetAuthorizationToken",
        #                 "s3:GetAccessPoint",
        #                 "s3:GetAccountPublicAccessBlock",
        #                 "s3:ListAccessPoints",
        #                 "s3:ListJobs"
        #             ],
        #             "Resource": [
        #                 "*"
        #             ]
        #         },
        #         {
        #             "Sid": "S3PermissionsmanagementBucket",
        #             "Effect": "Allow",
        #             "Action": [
        #                 "s3:DeleteBucketPolicy",
        #                 "s3:PutBucketAcl",
        #                 "s3:PutBucketPolicy",
        #                 "s3:PutBucketPublicAccessBlock"
        #             ],
        #             "Resource": [
        #                 "arn:aws:s3:::example-org-s3-access-logs"
        #             ]
        #         }
        #     ]
        # }

        print(json.dumps(results, indent=4))
        self.maxDiff = None

        # To future-proof this unit test...
        # (1) check the Sid names
        sid_names = get_sid_names_from_policy(results)
        self.assertIn(
            "MultMultNone",
            sid_names,
            "Sid is not in the list of expected Statement Ids",
        )
        self.assertIn(
            "S3PermissionsmanagementBucket",
            sid_names,
            "Sid is not in the list of expected Statement Ids",
        )

        # (2) Check for the presence of certain actions that we know will be there
        statement_1 = get_statement_from_policy_using_sid(results, "MultMultNone")
        self.assertIn("ram:EnableSharingWithAwsOrganization", statement_1.get("Action"))
        self.assertIn("s3:GetAccountPublicAccessBlock", statement_1.get("Action"))

        statement_2 = get_statement_from_policy_using_sid(results, "S3PermissionsmanagementBucket")
        self.assertIn("s3:DeleteBucketPolicy", statement_2.get("Action"))
        self.assertIn("s3:PutBucketPolicy", statement_2.get("Action"))

        # (3) Check that the length of the list is at least the length that it is right now,
        # since we expect it to grow eventually
        self.assertTrue(len(statement_1.get("Action")) > 5)  # Size is 6 at time of writing
        self.assertTrue(len(statement_2.get("Action")) > 3)  # Size is 4 at time of writing

    def test_eks_gh_155(self):
        """test_eks_gh_155: Test EKS issue raised in GH-155"""
        template_file_path = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                os.path.pardir,
                "files",
                "eks-service-wide.yml",
            )
        )
        cfg = read_yaml_file(template_file_path)
        result = write_policy_with_template(cfg)
        print(json.dumps(result, indent=4))
        expected_results = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "MultMultNone",
                    "Effect": "Allow",
                    "Action": [
                        "eks:ListClusters",
                        "eks:CreateCluster",
                        "eks:RegisterCluster",
                    ],
                    "Resource": ["*"],
                }
            ],
        }
        self.assertEqual(result["Statement"][0]["Sid"], expected_results["Statement"][0]["Sid"])
        self.assertTrue("eks:ListClusters" in result["Statement"][0]["Action"])
        self.assertTrue("eks:CreateCluster" in result["Statement"][0]["Action"])
        self.assertTrue("eks:RegisterCluster" in result["Statement"][0]["Action"])

    def test_dynamodb_arn_policy_gh_215(self):
        """test_dynamodb_arn_matching_gh_215: Test writing a policy with DynamoDB"""
        template_file_path = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                os.path.pardir,
                "files",
                "dynamodb_gh_215.yml",
            )
        )
        cfg = read_yaml_file(template_file_path)
        results = write_policy_with_template(cfg)
        print(json.dumps(results, indent=4))
        expected_statement_ids = [
            "MultMultNone",
            "DynamodbReadTable",
            "DynamodbWriteTable",
        ]
        for statement in results.get("Statement"):
            self.assertTrue(statement.get("Sid") in expected_statement_ids)


class RdsWritingTestCase(unittest.TestCase):
    def test_rds_policy_read_only(self):
        """test_rds_policy_read_only: Make sure that the RDS Policies work properly"""
        policy_file_path = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                os.path.pardir,
                os.path.pardir,
                "examples",
                "yml",
                "crud-rds-read.yml",
            )
        )
        cfg = read_yaml_file(policy_file_path)
        desired_output = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "RdsReadDb",
                    "Effect": "Allow",
                    "Action": [
                        "rds:DownloadCompleteDBLogFile",
                        "rds:DownloadDBLogFilePortion",
                        "rds:ListTagsForResource",
                    ],
                    "Resource": ["arn:aws:rds:us-east-1:123456789012:db:mydatabase"],
                }
            ],
        }
        expected_actions = [
            "rds:DownloadCompleteDBLogFile",
            "rds:DownloadDBLogFilePortion",
            "rds:ListTagsForResource",
        ]
        output = write_policy_with_template(cfg)
        print(json.dumps(output, indent=4))
        expected_statement_ids = ["RdsReadDb"]
        for statement in output.get("Statement"):
            self.assertTrue(statement.get("Sid") in expected_statement_ids)
        for action in expected_actions:
            self.assertTrue(action in output["Statement"][0]["Action"])

    def test_rds_policy_read_write_list(self):
        """test_rds_policy_read_write_list: Make sure that the RDS Policies work properly for multiple levels"""
        policy_file_path = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                os.path.pardir,
                os.path.pardir,
                "examples",
                "yml",
                "crud-rds-mult.yml",
            )
        )
        cfg = read_yaml_file(policy_file_path)
        desired_output = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "RdsReadDb",
                    "Effect": "Allow",
                    "Action": [
                        "rds:DownloadDBLogFilePortion",
                        "rds:ListTagsForResource",
                    ],
                    "Resource": ["arn:aws:rds:us-east-1:123456789012:db:mydatabase"],
                },
                {
                    "Sid": "MultMultNone",
                    "Effect": "Allow",
                    "Action": ["iam:PassRole"],
                    "Resource": ["*"],
                },
                {
                    "Sid": "RdsWriteDb",
                    "Effect": "Allow",
                    "Action": [
                        "rds:AddRoleToDBInstance",
                        "rds:ApplyPendingMaintenanceAction",
                        "rds:CreateDBInstance",
                        "rds:CreateDBInstanceReadReplica",
                        "rds:CreateDBSnapshot",
                        "rds:DeleteDBInstance",
                        "rds:DeregisterDBProxyTargets",
                        "rds:ModifyDBInstance",
                        "rds:PromoteReadReplica",
                        "rds:RebootDBInstance",
                        "rds:RemoveRoleFromDBInstance",
                        "rds:RestoreDBInstanceFromDBSnapshot",
                        "rds:RestoreDBInstanceFromS3",
                        "rds:RestoreDBInstanceToPointInTime",
                        "rds:StartDBInstance",
                        "rds:StopDBInstance",
                    ],
                    "Resource": ["arn:aws:rds:us-east-1:123456789012:db:mydatabase"],
                },
                {
                    "Sid": "RdsListDb",
                    "Effect": "Allow",
                    "Action": [
                        "rds:DescribeDBLogFiles",
                        "rds:DescribeDBProxyTargets",
                        "rds:DescribeDBSnapshots",
                        "rds:DescribePendingMaintenanceActions",
                        "rds:DescribeValidDBInstanceModifications",
                    ],
                    "Resource": ["arn:aws:rds:us-east-1:123456789012:db:mydatabase"],
                },
            ],
        }
        expected_statement_ids = [
            "RdsReadDb",
            "MultMultNone",
            "RdsWriteDb",
            "RdsListDb",
        ]
        policy = write_policy_with_template(cfg)
        for statement in policy.get("Statement"):
            self.assertTrue(statement.get("Sid") in expected_statement_ids)
        # print(json.dumps(policy, indent=4))
        # self.assertDictEqual(desired_output, policy)



import os
import unittest
import json
from policy_sentry.util.policy_files import (
    get_actions_from_json_policy_file,
    get_actions_from_policy,
    get_statement_from_policy_using_sid,
    get_sid_names_from_policy,
)


class PolicyFilesTestCase(unittest.TestCase):
    def test_get_actions_from_policy(self):
        """util.policy_files.get_actions_from_policy"""
        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": [
                        "ecr:getauthorizationtoken",
                        "ecr:batchchecklayeravailability",
                        "ecr:getdownloadurlforlayer",
                        "ecr:getrepositorypolicy",
                        "ecr:describerepositories",
                        "ecr:listimages",
                        "ecr:describeimages",
                        "ecr:batchgetimage",
                        "ecr:initiatelayerupload",
                        "ecr:uploadlayerpart",
                        "ecr:completelayerupload",
                        "ecr:putimage",
                    ],
                    "Resource": "*",
                },
                {
                    "Sid": "AllowManageOwnAccessKeys",
                    "Effect": "Allow",
                    "Action": [
                        "iam:CreateAccessKey",
                        "iam:DeleteAccessKey",
                        "iam:ListAccessKeys",
                        "iam:UpdateAccessKey",
                    ],
                    "Resource": "arn:aws:iam::*:user/${aws:username}",
                },
            ],
        }
        actions_list = get_actions_from_policy(policy)
        desired_actions_list = [
            "ecr:BatchCheckLayerAvailability",
            "ecr:BatchGetImage",
            "ecr:CompleteLayerUpload",
            "ecr:DescribeImages",
            "ecr:DescribeRepositories",
            "ecr:GetAuthorizationToken",
            "ecr:GetDownloadUrlForLayer",
            "ecr:GetRepositoryPolicy",
            "ecr:InitiateLayerUpload",
            "ecr:ListImages",
            "ecr:PutImage",
            "ecr:UploadLayerPart",
            "iam:CreateAccessKey",
            "iam:DeleteAccessKey",
            "iam:ListAccessKeys",
            "iam:UpdateAccessKey",
        ]
        self.maxDiff = None
        print(json.dumps(actions_list, indent=4))
        self.assertListEqual(desired_actions_list, actions_list)

    def test_get_actions_from_policy_file_with_explicit_actions(self):
        """util.policy_file.get_actions_from_policy_file_with_explicit_actions: Verify that we can get a list of actions from a
        file when it contains specific actions"""
        policy_file_path = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                os.path.pardir,
                os.path.pardir,
                "examples",
                "analyze",
                "explicit-actions.json",
            )
        )
        requested_actions = get_actions_from_json_policy_file(policy_file_path)
        # print(requested_actions)
        self.maxDiff = None
        print(json.dumps(requested_actions, indent=4))
        desired_actions = [
            "ecr:BatchCheckLayerAvailability",
            "ecr:BatchGetImage",
            "ecr:CompleteLayerUpload",
            "ecr:DescribeImages",
            "ecr:DescribeRepositories",
            "ecr:GetAuthorizationToken",
            "ecr:GetDownloadUrlForLayer",
            "ecr:GetRepositoryPolicy",
            "ecr:InitiateLayerUpload",
            "ecr:ListImages",
            "ecr:PutImage",
            "ecr:UploadLayerPart",
            "iam:CreateAccessKey",
            "iam:DeleteAccessKey",
            "iam:ListAccessKeys",
            "iam:UpdateAccessKey",
        ]
        self.assertListEqual(requested_actions, desired_actions)

    def test_get_sid_names_from_policy(self):
        """util.policy_files.get_sid_names_from_policy"""
        policy_json = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "MultMultNone",
                    "Effect": "Allow",
                    "Action": [
                        "ram:EnableSharingWithAwsOrganization",
                        "ram:GetResourcePolicies",
                        "ecr:GetAuthorizationToken",
                        "s3:GetAccessPoint",
                        "s3:GetAccountPublicAccessBlock",
                        "s3:ListAccessPoints",
                    ],
                    "Resource": ["*"],
                },
                {
                    "Sid": "S3PermissionsmanagementBucket",
                    "Effect": "Allow",
                    "Action": [
                        "s3:DeleteBucketPolicy",
                        "s3:PutBucketAcl",
                        "s3:PutBucketPolicy",
                        "s3:PutBucketPublicAccessBlock",
                    ],
                    "Resource": ["arn:aws:s3:::example-org-s3-access-logs"],
                },
            ],
        }
        expected_result = ["MultMultNone", "S3PermissionsmanagementBucket"]
        result = get_sid_names_from_policy(policy_json)
        self.assertListEqual(result, expected_result)

    def test_get_statement_from_policy_using_sid(self):
        """util.policy_files.get_statement_from_policy_using_sid"""
        policy_json = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "MultMultNone",
                    "Effect": "Allow",
                    "Action": [
                        "ram:EnableSharingWithAwsOrganization",
                        "ram:GetResourcePolicies",
                        "ecr:GetAuthorizationToken",
                        "s3:GetAccessPoint",
                        "s3:GetAccountPublicAccessBlock",
                        "s3:ListAccessPoints",
                    ],
                    "Resource": ["*"],
                },
                {
                    "Sid": "S3PermissionsmanagementBucket",
                    "Effect": "Allow",
                    "Action": [
                        "s3:DeleteBucketPolicy",
                        "s3:PutBucketAcl",
                        "s3:PutBucketPolicy",
                        "s3:PutBucketPublicAccessBlock",
                    ],
                    "Resource": ["arn:aws:s3:::example-org-s3-access-logs"],
                },
            ],
        }

        expected_result = {
            "Sid": "MultMultNone",
            "Effect": "Allow",
            "Action": [
                "ram:EnableSharingWithAwsOrganization",
                "ram:GetResourcePolicies",
                "ecr:GetAuthorizationToken",
                "s3:GetAccessPoint",
                "s3:GetAccountPublicAccessBlock",
                "s3:ListAccessPoints",
            ],
            "Resource": ["*"],
        }

        result = get_statement_from_policy_using_sid(policy_json, "MultMultNone")
        self.assertDictEqual(result, expected_result)



"""
Just some text transformation functions related to our
conditions-related workarounds
"""


# Borrowed from parliament. Altered with normalization for lowercase.
def translate_condition_key_data_types(condition_str: str) -> str:
    """
    The docs use different type names, so this standardizes them.
    Example: The condition secretsmanager:RecoveryWindowInDays is listed as using a "Long"
    So return "Number"
    """
    condition_lowercase = condition_str.lower()
    if condition_lowercase in ("arn", "arn"):
        return "Arn"
    if condition_lowercase in ("bool", "boolean"):
        return "Bool"
    if condition_lowercase == "date":
        return "Date"
    if condition_lowercase in ("long", "numeric"):
        return "Number"
    if condition_lowercase in ("string", "string", "arrayofstring"):
        return "String"
    if condition_lowercase == "ip":
        return "Ip"

    raise Exception(f"Unknown data format: {condition_lowercase}")


def get_service_from_condition_key(condition_key: str) -> str:
    """Given a condition key, return the service prefix"""
    return condition_key.split(":", 2)[0]


def get_comma_separated_condition_keys(condition_keys: str) -> str:
    """
    :param condition_keys: String containing multiple condition keys, separated by double spaces
    :return: result: String containing multiple condition keys, comma-separated
    """

    # replace the double spaces with a comma
    return condition_keys.replace("  ", ",")


def is_condition_key_match(document_key: str, some_str: str) -> bool:
    """Given a documented condition key and one from a policy, determine if they match
    Examples:
    - s3:prefix and s3:prefix obviously match
    - s3:ExistingObjectTag/<key> and s3:ExistingObjectTag/backup match
    """

    # Normalize both
    document_key = document_key.lower()
    some_str = some_str.lower()

    # Check if the document key has a pattern match in it
    if "$" in document_key:
        # Some services use a format like license-manager:ResourceTag/${TagKey}
        if some_str.startswith(document_key.split("$")[0]):
            return True
    elif "<" in document_key:
        # Some services use a format like s3:ExistingObjectTag/<key>
        if some_str.startswith(document_key.split("<")[0]):
            return True
    elif "tag-key" in document_key:  # noqa: SIM102
        # Some services use a format like secretsmanager:ResourceTag/tag-key
        if some_str.startswith(document_key.split("tag-key")[0]):
            return True

    # Just return whether they match exactly
    return document_key == some_str



"""
Text operations specific to IAM actions
"""

from __future__ import annotations


def get_service_from_action(action: str) -> str:
    """
    Returns the service name from a service:action combination
    :param action: ec2:DescribeInstance
    :return: ec2
    """
    service = action.split(":")[0]
    return service.lower()


def get_action_name_from_action(action: str) -> str:
    """
    Returns the lowercase action name from a service:action combination
    :param action: ec2:DescribeInstance
    :return: describeinstance
    """
    action_name = action.split(":")[1]
    return action_name.lower()


def get_lowercase_action_list(action_list: list[str]) -> list[str]:
    """
    Given a list of actions, return the list but in lowercase format
    """
    return [action.lower() for action in action_list]


def get_full_action_name(service: str, action_name: str) -> str:
    """
    Gets the proper formatting for an action - the service, plus colon, plus action name.
    :param service: service name, like s3
    :param action_name: action name, like createbucket
    :return: the resulting string
    """
    return f"{service}:{action_name}"



"""
Functions to use for parsing ARNs, matching ARN types, and getting the right fragment/component from an ARN string,
"""

#     Case 1: arn:partition:service:region:account-id:resource
#     Case 2: arn:partition:service:region:account-id:resourcetype/resource
#     Case 3: arn:partition:service:region:account-id:resourcetype/resource/qualifier
#     Case 4: arn:partition:service:region:account-id:resourcetype/resource:qualifier
#     Case 5: arn:partition:service:region:account-id:resourcetype:resource
#     Case 6: arn:partition:service:region:account-id:resourcetype:resource:qualifier
#     Source: https://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html#genref-arns
from __future__ import annotations

import logging
import re

ARN_SEPARATOR_PATTERN = re.compile(r"[:/]")
# Note: Each service can only have one of these, so these are definitely exceptions
EXCLUSION_LIST = {
    "${ObjectName}",
    "${RepositoryName}",
    "${BucketName}",
    "table/${TableName}",
    "${BucketName}/${ObjectName}",
}

logger = logging.getLogger(__name__)


# pylint: disable=too-many-instance-attributes
class ARN:
    """Class that helps to match ARN resource type formats neatly"""

    def __init__(self, provided_arn: str) -> None:
        self.arn = provided_arn
        follows_arn_format = re.search(r"^arn:([^:]*):([^:]*):([^:]*):([^:]*):(.+)$", provided_arn)

        if not follows_arn_format:
            raise Exception("The provided value does not follow required ARN format.")
        try:
            elements = self.arn.split(":", 5)
            self.partition = elements[1]
            self.service_prefix = elements[2]
            self.region = elements[3]
            self.account = elements[4]
            self.resource = elements[5]
        except IndexError as error:
            raise Exception(f"The provided ARN is invalid. IndexError: {error}. Please provide a valid ARN.") from error
        if "/" in self.resource:
            self.resource, self.resource_path = self.resource.split("/", 1)
        elif ":" in self.resource:
            self.resource, self.resource_path = self.resource.split(":", 1)
        self.resource_string = self._resource_string()

    def __repr__(self) -> str:
        return self.arn

    # pylint: disable=too-many-return-statements
    def _resource_string(self) -> str:
        """
        Given an ARN, return the string after the account ID, no matter the ARN format.
        Return:
            String: The resource string, like `resourcetype/resource`
        """
        # Resource string will equal:
        #     Case 1: resource
        #     Case 2: resourcetype/resource
        #     Case 3: resourcetype/resource/qualifier
        #     Case 4: resourcetype/resource:qualifier
        #     Case 5: resourcetype:resource
        #     Case 6: resourcetype:resource:qualifier
        split_arn = self.arn.split(":")
        return ":".join(split_arn[5:])

    def same_resource_type(self, arn_in_database: str) -> bool:
        """Given an arn, see if it has the same resource type"""

        # 1. If the RAW ARN in the database is *, then it doesn't have a resource type
        if arn_in_database == "*":
            return False

        # 2. ARNs should have the same service
        elements = arn_in_database.split(":", 5)
        if self.service_prefix != elements[2]:
            return False

        # 3. Add support for resource type wildcards, per #204
        # When specifying an ARN and setting a * on the resource type (e.g. instead of db:instance it is *:*),
        #   multiple ARNs can match.
        #   Previously, this would fail and return empty results.
        #   Now it correctly returns the full list of matching ARNs and corresponding actions.
        resource_type_arn_to_test = parse_arn_for_resource_type(self.arn)
        if resource_type_arn_to_test == "*":
            return True

        # 4. Match patterns for complicated resource strings, leveraging the standardized format of the Raw ARN format
        # table/${TableName} should not match `table/${TableName}/backup/${BackupName}`
        resource_string_arn_in_database = get_resource_string(arn_in_database)

        split_resource_string_in_database = re.split(ARN_SEPARATOR_PATTERN, resource_string_arn_in_database)
        # logger.debug(str(split_resource_string_in_database))
        arn_format_list = []
        for elem in split_resource_string_in_database:
            if "${" not in elem:
                arn_format_list.append(elem)
            else:
                # If an element says something like ${TableName}, normalize it to an empty string
                arn_format_list.append("")

        split_resource_string_to_test = re.split(ARN_SEPARATOR_PATTERN, self.resource_string)
        # 4b: If we have a confusing resource string, the length of the split resource string list
        #  should at least be the same
        # Again, table/${TableName} (len of 2) should not match `table/${TableName}/backup/${BackupName}` (len of 4)
        # if len(split_resource_string_to_test) != len(arn_format_list):
        #     return False

        lower_resource_string = [x.lower() for x in split_resource_string_to_test]
        for elem in arn_format_list:
            if elem and elem.lower() not in lower_resource_string:
                return False

        # 4c: See if the non-normalized fields match
        for idx, elem in enumerate(arn_format_list):
            # If the field is not normalized to empty string, then make sure the resource type segments match
            # So, using table/${TableName}/backup/${BackupName} as an example:
            # table should match, backup should match,
            # and length of the arn_format_list should be the same as split_resource_string_to_test
            # If all conditions match, then the ARN format is the same.
            if elem:
                if elem == split_resource_string_to_test[idx] or split_resource_string_to_test[idx] == "*":
                    pass
                else:
                    return False

        # 4. Special type for S3 bucket objects and CodeCommit repos
        resource_path_arn_in_database = elements[5]
        if resource_path_arn_in_database in EXCLUSION_LIST:
            logger.debug(f"Special type: {resource_path_arn_in_database}")
            # handling special case table/${TableName}
            if resource_string_arn_in_database in (
                "table/${TableName}",
                "${BucketName}",
            ):
                return len(self.resource_string.split("/")) == len(elements[5].split("/"))
            # If we've made it this far, then it is a special type
            # return True
            # Presence of / would mean it's an object in both so it matches
            if "/" in self.resource_string and "/" in elements[5]:
                return True
            # / not being present in either means it's a bucket in both so it matches
            if "/" not in self.resource_string and "/" not in elements[5]:  # noqa: SIM103
                return True
            # If there is a / in one but not in the other, it does not match
            return False

        # 5. If we've made it this far, then it should pass
        return True


def parse_arn(arn: str) -> dict[str, str]:
    """
    Given an ARN, split up the ARN into the ARN namespacing schema dictated by the AWS docs.
    """
    try:
        elements = arn.split(":", 5)
        result = {
            "arn": elements[0],
            "partition": elements[1],
            "service": elements[2],
            "region": elements[3],
            "account": elements[4],
            "resource": elements[5],
            "resource_path": "",
        }
    except IndexError as error:
        raise Exception(f"IndexError: The provided ARN '{arn}' is invalid. Please provide a valid ARN.") from error
    if "/" in result["resource"]:
        result["resource"], result["resource_path"] = result["resource"].split("/", 1)
    elif ":" in result["resource"]:
        result["resource"], result["resource_path"] = result["resource"].split(":", 1)
    return result


def get_service_from_arn(arn: str) -> str:
    """Given an ARN string, return the service"""
    result = parse_arn(arn)
    return result["service"]


def get_region_from_arn(arn: str) -> str:
    """Given an ARN, return the region in the ARN, if it is available. In certain cases like S3 it is not"""
    result = parse_arn(arn)
    # Support S3 buckets with no values under region
    if result["region"] is None:
        return ""
    return result["region"]


def get_account_from_arn(arn: str) -> str:
    """Given an ARN, return the account ID in the ARN, if it is available. In certain cases like S3 it is not"""
    result = parse_arn(arn)
    # Support S3 buckets with no values under account
    if result["account"] is None:
        return ""
    return result["account"]


def get_resource_path_from_arn(arn: str) -> str | None:
    """Given an ARN, parse it according to ARN namespacing and return the resource path. See
    http://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html for more details on ARN namespacing.
    """
    result = parse_arn(arn)
    return result["resource_path"]


def get_resource_string(arn: str) -> str:
    """
    Given an ARN, return the string after the account ID, no matter the ARN format.

    Arguments:
        arn: An ARN, like `arn:partition:service:region:account-id:resourcetype/resource`
    Return:
        String: The resource string, like `resourcetype/resource`
    """
    split_arn = arn.split(":")
    return ":".join(split_arn[5:])


# In the meantime, we have to skip this pylint check (consider this as tech debt)
# pylint: disable=inconsistent-return-statements
def parse_arn_for_resource_type(arn: str) -> str | None:
    """
    Parses the resource string (resourcetype/resource and other variants) and grab the resource type.

    Arguments:
        arn: The resource string to parse, like `resourcetype/resource`
    Return:
        String: The resource type, like `bucket` or `object`
    """
    split_arn = arn.split(":")
    resource_string = ":".join(split_arn[5:])
    split_resource = re.split(ARN_SEPARATOR_PATTERN, resource_string)
    if len(split_resource) == 1:
        # logger.debug(f"split_resource length is 1: {str(split_resource)}")
        pass
    elif len(split_resource) > 1:
        return split_resource[0]

    return None


def does_arn_match(arn_to_test: str, arn_in_database: str) -> bool:
    """
    Given two ARNs, determine if they have the same resource type.

    Arguments:
        arn_to_test: ARN provided by user
        arn_in_database: Raw ARN that exists in the policy sentry database

    Returns:
        Boolean: result of whether or not the ARNs match
    """
    this_arn = ARN(arn_to_test)
    return this_arn.same_resource_type(arn_in_database)



"""Used for loading the IAM data"""

from __future__ import annotations

import functools
import gc
import logging
from typing import Any, cast

import orjson

from policy_sentry.shared.constants import (
    DATASTORE_FILE_PATH,
    POLICY_SENTRY_SCHEMA_VERSION_NAME,
)

logger = logging.getLogger()
# On initialization, load the IAM data
iam_definition_path = DATASTORE_FILE_PATH
logger.debug(f"Leveraging the IAM definition at {iam_definition_path}")


def load_iam_definition() -> dict[str, Any]:
    gc_enabled = gc.isenabled()
    if gc_enabled:
        # https://github.com/msgpack/msgpack-python?tab=readme-ov-file#performance-tips
        gc.disable()

    data: dict[str, Any] = orjson.loads(iam_definition_path.read_bytes())

    if gc_enabled:
        gc.enable()

    return data


iam_definition = load_iam_definition()


@functools.lru_cache(maxsize=1)
def get_iam_definition_schema_version() -> str:
    """
    Returns the schema version of the IAM datastore
    """

    return cast(
        "str",
        iam_definition[POLICY_SENTRY_SCHEMA_VERSION_NAME],
    )


@functools.lru_cache(maxsize=1024)
def get_service_prefix_data(service_prefix: str) -> dict[str, Any]:
    """
    Given an AWS service prefix, return a large dictionary of IAM privilege data for processing and analysis.

    Arguments:
        service_prefix: An AWS service prefix, like s3 or ssm
    Returns:
        List: A list of metadata about that service
    """
    try:
        return cast("dict[str, Any]", iam_definition[service_prefix])
    except:  # noqa: E722
        if service_prefix == "catalog":
            # the resource types "Portfolio" and "Product" have the service name "catalog" in their ARN
            # https://docs.aws.amazon.com/service-authorization/latest/reference/list_awsservicecatalog.html#awsservicecatalog-resources-for-iam-policies
            return cast("dict[str, Any]", iam_definition["servicecatalog"])

        logger.info(f"Service prefix not {service_prefix} found.")
        return {}



import unittest
import json
from policy_sentry.querying.conditions import (
    get_condition_keys_for_service,
    get_condition_key_details,
    get_condition_keys_available_to_raw_arn,
    get_conditions_for_action_and_raw_arn,
    get_condition_value_type,
)


class QueryConditionsTestCase(unittest.TestCase):
    def test_get_condition_keys_for_service(self):
        """querying.conditions.get_condition_keys_for_service test"""
        expected_results = [
            "aws:RequestTag/${TagKey}",
            "aws:ResourceTag/${TagKey}",
            "aws:TagKeys",
            "ram:AllowsExternalPrincipals",
            "ram:PermissionArn",
            "ram:PermissionResourceType",
            "ram:Principal",
            "ram:RequestedAllowsExternalPrincipals",
            "ram:RequestedResourceType",
            "ram:ResourceArn",
            "ram:ResourceShareName",
            "ram:ShareOwnerAccountId",
        ]
        results = get_condition_keys_for_service("ram")
        for expected_result in expected_results:
            self.assertTrue(expected_result in results)
        print(results)
        # print(json.dumps(results, indent=4))
        # self.assertEqual(results, expected_results)

    def test_get_condition_keys_available_to_raw_arn(self):
        expected_results = [
            "aws:RequestTag/${TagKey}",
            "aws:ResourceTag/${TagKey}",
            "aws:TagKeys",
            "ec2:Region",
            "ec2:ResourceTag/${TagKey}",
            "ec2:Vpc",
        ]
        raw_arn = "arn:${Partition}:ec2:${Region}:${Account}:security-group/${SecurityGroupId}"
        result = get_condition_keys_available_to_raw_arn(raw_arn)
        print(result)
        for expected_result in expected_results:
            self.assertTrue(expected_result in result)
        # self.assertListEqual(expected_results, result)

    def test_get_condition_key_details(self):
        """querying.conditions.get_condition_key_details"""
        desired_output = {
            "name": "cloud9:Permissions",
            "description": "Filters access by the type of AWS Cloud9 permissions",
            "condition_value_type": "string",
        }
        output = get_condition_key_details("cloud9", "cloud9:Permissions")
        self.assertEqual(desired_output, output)

    def test_get_conditions_for_action_and_raw_arn(self):
        """querying.conditions.get_conditions_for_action_and_raw_arn"""
        desired_condition_keys_list = [
            "aws:RequestTag/${TagKey}",
            "aws:ResourceTag/${TagKey}",
            "aws:TagKeys",
            "ec2:Region",
            "ec2:ResourceTag/${TagKey}",
            "ec2:Vpc",
        ]
        output = get_conditions_for_action_and_raw_arn(
            "ec2:AuthorizeSecurityGroupEgress",
            "arn:${Partition}:ec2:${Region}:${Account}:security-group/${SecurityGroupId}",
        )
        self.maxDiff = None
        # print(output)
        for condition in desired_condition_keys_list:
            self.assertTrue(condition in output)
        # self.assertListEqual(desired_condition_keys_list, output)

    def test_get_condition_value_type(self):
        """querying.conditions.get_condition_value_type"""
        desired_result = "arn"
        condition_key = "secretsmanager:SecretId"
        result = get_condition_value_type(condition_key)
        self.maxDiff = None
        # print(result)
        self.assertEqual(desired_result, result)

    def test_gh_225_s3_conditions(self):
        """querying.actions.get_actions_matching_condition_key"""
        results = get_condition_keys_for_service("s3")
        print(json.dumps(results, indent=4))
        expected_results = [
            "aws:RequestTag/${TagKey}",
            "aws:ResourceTag/${TagKey}",
            "aws:TagKeys",
            "s3:AccessPointNetworkOrigin",
            "s3:DataAccessPointAccount",
            "s3:DataAccessPointArn",
            "s3:ExistingJobOperation",
            "s3:ExistingJobPriority",
            "s3:ExistingObjectTag/<key>",
            "s3:JobSuspendedCause",
            "s3:LocationConstraint",
            "s3:RequestJobOperation",
            "s3:RequestJobPriority",
            "s3:RequestObjectTag/<key>",
            "s3:RequestObjectTagKeys",
            "s3:VersionId",
            "s3:authType",
            "s3:delimiter",
            "s3:locationconstraint",
            "s3:max-keys",
            "s3:object-lock-legal-hold",
            "s3:object-lock-mode",
            "s3:object-lock-remaining-retention-days",
            "s3:object-lock-retain-until-date",
            "s3:prefix",
            "s3:signatureAge",
            "s3:signatureversion",
            "s3:versionid",
            "s3:x-amz-acl",
            "s3:x-amz-content-sha256",
            "s3:x-amz-copy-source",
            "s3:x-amz-grant-full-control",
            "s3:x-amz-grant-read",
            "s3:x-amz-grant-read-acp",
            "s3:x-amz-grant-write",
            "s3:x-amz-grant-write-acp",
            "s3:x-amz-metadata-directive",
            "s3:x-amz-server-side-encryption",
            "s3:x-amz-server-side-encryption-aws-kms-key-id",
            "s3:x-amz-storage-class",
            "s3:x-amz-website-redirect-location",
        ]


#         for expected_result in expected_results:
#             self.assertTrue(expected_result in results)
# self.assertListEqual(results, expected_results)



import unittest
from policy_sentry.writing.validate import check_actions_schema, check_crud_schema

valid_cfg_for_crud = {
    "mode": "crud",
    "name": "RoleNameWithCRUD",
    "read": [
        "arn:aws:s3:::example-org-sbx-vmimport",
    ],
    "write": [
        "arn:aws:s3:::example-org-s3-access-logs",
    ],
    "list": [
        "arn:aws:s3:::example-org-flow-logs",
        "arn:aws:s3:::example-org-sbx-vmimport/stuff",
    ],
    "tagging": ["arn:aws:ssm:us-east-1:123456789012:parameter/test"],
    "permissions-management": ["arn:aws:s3:::example-org-s3-access-logs"],
    "wildcard-only": {"service-read": ["s3"]},
}

valid_cfg_for_actions = {
    "mode": "actions",
    "name": "RoleNameWithActions",
    "actions": [
        "kms:CreateGrant",
        "kms:CreateCustomKeyStore",
        "ec2:AuthorizeSecurityGroupEgress",
        "ec2:AuthorizeSecurityGroupIngress",
    ],
}

valid_crud_with_one_item_only = {
    "mode": "crud",
    "name": "RoleNameWithCRUD",
    "read": [
        "arn:aws:s3:::example-org-sbx-vmimport",
    ],
}

invalid_crud_with_mispelled_category = {
    "mode": "crud",
    "name": "RoleNameWithCRUD",
    "reed": [
        "arn:aws:s3:::example-org-sbx-vmimport",
    ],
    "right": [
        "arn:aws:s3:::example-org-sbx-vmimport",
    ],
    "wrist": [
        "arn:aws:s3:::example-org-sbx-vmimport",
    ],
}

valid_crud_with_sts_actions = {
    "mode": "crud",
    "name": "RoleNameWithCRUD",
    "sts": {
        "assume-role": [
            "arn:aws:iam::123456789012:role/demo",
        ]
    },
}

invalid_crud_with_invalid_sts_action = {
    "mode": "crud",
    "name": "RoleNameWithCRUD",
    "sts": {
        "assume-role-fake-something": [
            "arn:aws:iam::123456789012:role/demo",
        ]
    },
}


class YMLSchemaTestCase(unittest.TestCase):
    def test_actions_schema(self):
        """test_actions_schema: Validates that the user-supplied YAML is working for CRUD mode"""
        result = check_actions_schema(valid_cfg_for_actions)
        self.assertTrue(result)

    def test_crud_schema(self):
        """test_actions_schema: Validates that the user-supplied YAML is working for CRUD mode"""
        result = check_crud_schema(valid_cfg_for_crud)
        self.assertTrue(result)
        self.assertTrue(check_crud_schema(valid_crud_with_one_item_only))
        self.assertTrue(check_crud_schema(valid_crud_with_sts_actions))
        with self.assertRaises(Exception):
            check_crud_schema(invalid_crud_with_mispelled_category)
        with self.assertRaises(Exception):
            check_crud_schema(invalid_crud_with_invalid_sts_action)



import unittest
import json
from policy_sentry.querying.all import (
    get_all_service_prefixes,
    get_all_actions,
    get_service_authorization_url,
)
from policy_sentry.command.query import query_action_table


class QueryActionsTestCase(unittest.TestCase):
    def test_get_all_service_prefixes(self):
        result = get_all_service_prefixes()
        self.assertGreater(len(result), 212)
        # print(len(result))
        # print(json.dumps(result, indent=4))
        # performance notes:
        # old: 0.013s
        # new: 0.005s

    def test_get_all_actions(self):
        result = get_all_actions()
        print("Total count of unique IAM actions:", len(result))
        self.assertGreater(len(result), 7000)

        # performance notes:
        # old: 0.112s (without sort)
        # new: 0.106s

    def test_query_actions_with_access_level_and_wildcard_only(self):
        service = "all"
        resource_type = "*"
        result = query_action_table(
            service=service,
            resource_type=resource_type,
            name=None,
            access_level="permissions-management",
            condition=None,
        )
        print(len(result))
        self.assertTrue(len(result) > 200)

    def test_GH_296_query_all_actions_with_wildcard_resources(self):
        service = "all"
        resource_type = "*"
        result = query_action_table(
            service=service,
            resource_type=resource_type,
            name=None,
            access_level=None,
            condition=None,
        )
        self.assertGreater(len(result), 3000)

    def test_get_service_authorization_url(self):
        result = get_service_authorization_url("a4b")
        print(result)
        expected_result = (
            "https://docs.aws.amazon.com/service-authorization/latest/reference/list_alexaforbusiness.html"
        )
        self.assertTrue(result == expected_result)

    def test_query_catalog_service(self):
        # given
        service = "catalog"

        # when
        result = query_action_table(
            service=service,
            resource_type=None,
            name=None,
            access_level=None,
            condition=None,
        )

        # then
        self.assertFalse(result)



"""
Methods that execute specific queries against the SQLite database for the ARN table.
This supports the policy_sentry query functionality
"""

from __future__ import annotations

import functools
import logging
from typing import Any, cast

from policy_sentry.shared.iam_data import (
    get_service_prefix_data,
)
from policy_sentry.util.arns import does_arn_match, get_service_from_arn

logger = logging.getLogger(__name__)


@functools.lru_cache(maxsize=1024)
def get_raw_arns_for_service(service_prefix: str) -> list[str]:
    """
    Get a list of available raw ARNs per AWS service

    Arguments:
        service_prefix: An AWS service prefix, like `s3` or `kms`
    Returns:
        List: A list of raw ARNs
    """
    service_prefix_data = get_service_prefix_data(service_prefix)
    return [resource_data["arn"] for resource_data in service_prefix_data["resources"].values()]


@functools.lru_cache(maxsize=1024)
def get_arn_types_for_service(service_prefix: str) -> dict[str, str]:
    """
    Get a list of available ARN short names per AWS service.

    Arguments:
        service_prefix: An AWS service prefix, like `s3` or `kms`
    Returns:
        List: A list of ARN types, like `bucket` or `object`
    """
    service_prefix_data = get_service_prefix_data(service_prefix)
    return {
        resource_name: resource_data["arn"] for resource_name, resource_data in service_prefix_data["resources"].items()
    }


def get_arn_type_details(service_prefix: str, resource_type_name: str) -> dict[str, Any]:
    """
    Get details about ARNs in JSON format.

    Arguments:
        service_prefix: An AWS service prefix, like `s3` or `kms`
        resource_type_name: The name of a resource type, like `bucket` or `object`. To get details on ALL arns in a service, specify "*" here.
    Returns:
        Dictionary: Metadata about an ARN type
    """
    output = {}
    service_prefix_data = get_service_prefix_data(service_prefix)
    this_resource_type_name = service_prefix_data["resources_lower_name"].get(resource_type_name.lower())
    if this_resource_type_name:
        resource_data = service_prefix_data["resources"][this_resource_type_name]
        output = {
            "resource_type_name": this_resource_type_name,
            "raw_arn": resource_data["arn"],
            "condition_keys": resource_data["condition_keys"],
        }

    return output


# pylint: disable=inconsistent-return-statements
def get_resource_type_name_with_raw_arn(raw_arn: str) -> str | None:
    """
    Given a raw ARN, return the resource type name as shown in the database.

    Arguments:
        raw_arn: The raw ARN stored in the database, like 'arn:${Partition}:s3:::${BucketName}'
    Returns:
        String: The resource type name, like bucket
    """
    elements = raw_arn.split(":", 5)
    service_prefix = elements[2]
    service_data = get_service_prefix_data(service_prefix)

    for resource_name, resource_data in service_data["resources"].items():
        if resource_data["arn"].lower() == raw_arn.lower():
            return cast("str", resource_name)

    return None


def get_matching_raw_arns(arn: str) -> list[str]:
    """
    Given a user-supplied ARN, return the list of raw_arns since that is used as a unique identifier throughout this library

    Arguments:
        arn: The user-supplied arn, like arn:aws:s3:::mybucket
    Returns:
        list(str): The list of raw ARNs stored in the database, like 'arn:${Partition}:s3:::${BucketName}'
    """
    result = []
    service_in_scope = get_service_from_arn(arn)
    # Determine which resource it applies to
    all_raw_arns_for_service = get_raw_arns_for_service(service_in_scope)
    # Get the raw ARN specific to the provided one
    for raw_arn in all_raw_arns_for_service:
        if does_arn_match(arn, raw_arn) and raw_arn not in result:
            result.append(raw_arn)
    return result



import unittest
import json
import os
import yaml
from policy_sentry.writing.sid_group import SidGroup


crud_with_override_template = os.path.abspath(
    os.path.join(
        os.path.dirname(__file__),
        os.path.pardir,
        os.path.pardir,
        "examples",
        "yml",
        "crud-with-override.yml",
    )
)

with open(crud_with_override_template) as yaml_file:
    crud_with_override_template_cfg = yaml.safe_load(yaml_file)


class SidGroupCrudTestCase(unittest.TestCase):
    def test_sid_group_multiple(self):
        sid_group = SidGroup()
        arn_list_from_user = [
            "arn:aws:s3:::example-org-s3-access-logs",
            "arn:aws:kms:us-east-1:123456789012:key/123456",
        ]
        access_level = "Permissions management"
        sid_group.add_by_arn_and_access_level(arn_list_from_user, access_level)
        output = sid_group.get_sid_group()
        # print(json.dumps(output, indent=4))
        desired_output = {
            "S3PermissionsmanagementBucket": {
                "arn": ["arn:aws:s3:::example-org-s3-access-logs"],
                "service": "s3",
                "access_level": "Permissions management",
                "arn_format": "arn:${Partition}:s3:::${BucketName}",
                "actions": [
                    "s3:DeleteBucketPolicy",
                    "s3:PutBucketAcl",
                    "s3:PutBucketOwnershipControls",
                    "s3:PutBucketPolicy",
                    "s3:PutBucketPublicAccessBlock",
                ],
                "conditions": [],
            },
            "KmsPermissionsmanagementKey": {
                "arn": ["arn:aws:kms:us-east-1:123456789012:key/123456"],
                "service": "kms",
                "access_level": "Permissions management",
                "arn_format": "arn:${Partition}:kms:${Region}:${Account}:key/${KeyId}",
                "actions": [
                    "kms:CreateGrant",
                    "kms:PutKeyPolicy",
                    "kms:RetireGrant",
                    "kms:RevokeGrant",
                ],
                "conditions": [],
            },
        }
        self.maxDiff = None
        self.assertDictEqual(desired_output, output)
        desired_policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "S3PermissionsmanagementBucket",
                    "Effect": "Allow",
                    "Action": [
                        "s3:DeleteBucketPolicy",
                        "s3:PutBucketAcl",
                        "s3:PutBucketOwnershipControls",
                        "s3:PutBucketPolicy",
                        "s3:PutBucketPublicAccessBlock",
                    ],
                    "Resource": ["arn:aws:s3:::example-org-s3-access-logs"],
                },
                {
                    "Sid": "KmsPermissionsmanagementKey",
                    "Effect": "Allow",
                    "Action": [
                        "kms:CreateGrant",
                        "kms:PutKeyPolicy",
                        "kms:RetireGrant",
                        "kms:RevokeGrant",
                    ],
                    "Resource": ["arn:aws:kms:us-east-1:123456789012:key/123456"],
                },
            ],
        }
        rendered_policy = sid_group.get_rendered_policy()
        self.assertDictEqual(desired_policy, rendered_policy)
        # print(json.dumps(rendered_policy, indent=4))

    def test_sid_group(self):
        desired_output = {
            "S3PermissionsmanagementBucket": {
                "arn": ["arn:aws:s3:::example-org-s3-access-logs"],
                "service": "s3",
                "access_level": "Permissions management",
                "arn_format": "arn:${Partition}:s3:::${BucketName}",
                "actions": [
                    "s3:DeleteBucketPolicy",
                    "s3:PutBucketAcl",
                    "s3:PutBucketOwnershipControls",
                    "s3:PutBucketPolicy",
                    "s3:PutBucketPublicAccessBlock",
                ],
                "conditions": [],
            }
        }
        sid_group = SidGroup()
        arn_list_from_user = ["arn:aws:s3:::example-org-s3-access-logs"]
        access_level = "Permissions management"
        sid_group.add_by_arn_and_access_level(arn_list_from_user, access_level)
        status = sid_group.get_sid_group()
        self.maxDiff = None
        # print(json.dumps(status, indent=4))
        self.assertEqual(status, desired_output)
        rendered_policy = sid_group.get_rendered_policy()
        desired_policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "S3PermissionsmanagementBucket",
                    "Effect": "Allow",
                    "Action": [
                        "s3:DeleteBucketPolicy",
                        "s3:PutBucketAcl",
                        "s3:PutBucketOwnershipControls",
                        "s3:PutBucketPolicy",
                        "s3:PutBucketPublicAccessBlock",
                    ],
                    "Resource": ["arn:aws:s3:::example-org-s3-access-logs"],
                }
            ],
        }
        # print(json.dumps(rendered_policy, indent=4))
        self.maxDiff = None
        self.assertDictEqual(desired_policy, rendered_policy)

    # def test_get_actions_data_service_wide(self):
    #     data = get_action_data("s3", "*")
    #     # print(data)

    def test_refactored_crud_policy(self):
        """test_refactored_crud_policy"""
        sid_group = SidGroup()
        sid_group.add_by_arn_and_access_level(
            ["arn:aws:secretsmanager:us-east-1:123456789012:secret:mysecret"],
            "Read",
        )
        sid_group.add_by_arn_and_access_level(["arn:aws:s3:::example-org-sbx-vmimport/stuff"], "Tagging")
        sid_group.add_by_arn_and_access_level(
            ["arn:aws:secretsmanager:us-east-1:123456789012:secret:mysecret"],
            "Write",
        )
        sid_group.add_by_arn_and_access_level(
            ["arn:aws:secretsmanager:us-east-1:123456789012:secret:anothersecret"],
            "Write",
        )
        sid_group.add_by_arn_and_access_level(
            ["arn:aws:kms:us-east-1:123456789012:key/123456"],
            "Permissions management",
        )
        sid_group.add_by_arn_and_access_level(["arn:aws:ssm:us-east-1:123456789012:parameter/test"], "List")

        result = sid_group.get_rendered_policy()
        expected_statement_ids = [
            "SecretsmanagerReadSecret",
            "S3TaggingObject",
            "SecretsmanagerWriteSecret",
            "KmsPermissionsmanagementKey",
            "SsmListParameter",
        ]
        # print(json.dumps(output, indent=4))
        self.maxDiff = None
        print(json.dumps(result, indent=4))
        for statement in result.get("Statement"):
            self.assertTrue(statement.get("Sid") in expected_statement_ids)

    def test_write_with_template(self):
        cfg = {
            "mode": "crud",
            "name": "RoleNameWithCRUD",
            "permissions-management": ["arn:aws:s3:::example-org-s3-access-logs"],
            "list": ["arn:aws:secretsmanager:us-east-1:123456789012:secret:anothersecret"],
        }
        sid_group = SidGroup()
        rendered_policy = sid_group.process_template(cfg)
        desired_output = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "S3PermissionsmanagementBucket",
                    "Effect": "Allow",
                    "Action": [
                        "s3:DeleteBucketPolicy",
                        "s3:PutBucketAcl",
                        "s3:PutBucketOwnershipControls",
                        "s3:PutBucketPolicy",
                        "s3:PutBucketPublicAccessBlock",
                    ],
                    "Resource": ["arn:aws:s3:::example-org-s3-access-logs"],
                }
            ],
        }
        # print(json.dumps(rendered_policy, indent=4))
        self.assertEqual(rendered_policy, desired_output)

    def test_resource_restriction_plus_dependent_action(self):
        """
        test_resource_restriction_plus_dependent_action
        """
        # Given iam:generateorganizationsaccessreport with resource constraint, make sure these are added:
        #  organizations:DescribePolicy,organizations:ListChildren,organizations:ListParents,
        #  organizations:ListPoliciesForTarget,organizations:ListRoots,organizations:ListTargetsForPolicy
        actions_test_data_1 = ["iam:generateorganizationsaccessreport"]
        sid_group = SidGroup()
        sid_group.add_by_list_of_actions(actions_test_data_1)
        output = sid_group.get_rendered_policy()
        # print(json.dumps(rendered_policy, indent=4))
        desired_output = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": [
                        "organizations:DescribePolicy",
                        "organizations:ListChildren",
                        "organizations:ListParents",
                        "organizations:ListPoliciesForTarget",
                        "organizations:ListRoots",
                        "organizations:ListTargetsForPolicy",
                    ],
                    "Resource": ["*"],
                    "Sid": "MultMultNone",
                },
                {
                    "Sid": "IamReadAccessreport",
                    "Effect": "Allow",
                    "Action": ["iam:GenerateOrganizationsAccessReport"],
                    "Resource": ["arn:${Partition}:iam::${Account}:access-report/${EntityPath}"],
                },
            ],
        }
        self.maxDiff = None
        # print(json.dumps(output, indent=4))
        self.assertDictEqual(output, desired_output)

    def test_resource_restriction_plus_dependent_action_simple_2(self):
        """
        test_resource_restriction_plus_dependent_action_simple_2
        """
        # Given iam:generateorganizationsaccessreport with resource constraint, make sure these are added:
        #  organizations:DescribePolicy,organizations:ListChildren,organizations:ListParents,
        #  organizations:ListPoliciesForTarget,organizations:ListRoots,organizations:ListTargetsForPolicy

        sid_group = SidGroup()
        sid_group.add_by_arn_and_access_level(["arn:aws:iam::000000000000:access-report/somepath"], "Read")
        output = sid_group.get_rendered_policy()
        desired_output = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "MultMultNone",
                    "Effect": "Allow",
                    "Action": [
                        "organizations:DescribePolicy",
                        "organizations:ListChildren",
                        "organizations:ListParents",
                        "organizations:ListPoliciesForTarget",
                        "organizations:ListRoots",
                        "organizations:ListTargetsForPolicy",
                    ],
                    "Resource": ["*"],
                },
                {
                    "Sid": "IamReadAccessreport",
                    "Effect": "Allow",
                    "Action": ["iam:GenerateOrganizationsAccessReport"],
                    "Resource": ["arn:aws:iam::000000000000:access-report/somepath"],
                },
            ],
        }
        # print(json.dumps(output, indent=4))
        self.assertDictEqual(output, desired_output)

    def test_add_by_arn_and_access_level_ignores_placeholder_services(self):
        # https://github.com/salesforce/policy_sentry/issues/448

        sid_group = SidGroup()
        sid_group.add_by_arn_and_access_level(
            ["arn:${Partition}:${Vendor}:${Region}:*:${ResourceType}:${RecoveryPointId}"],
            "Read",
        )

        self.assertTrue(not sid_group.sids)

    def test_add_by_list_of_actions(self):
        actions_test_data_1 = ["kms:CreateCustomKeyStore", "kms:CreateGrant"]
        sid_group = SidGroup()
        sid_group.add_by_list_of_actions(actions_test_data_1)
        output = sid_group.get_rendered_policy()
        desired_output = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "KmsPermissionsmanagementKey",
                    "Effect": "Allow",
                    "Action": ["kms:CreateGrant"],
                    "Resource": ["arn:${Partition}:kms:${Region}:${Account}:key/${KeyId}"],
                },
                {
                    "Sid": "MultMultNone",
                    "Effect": "Allow",
                    "Action": [
                        "cloudhsm:DescribeClusters",
                        "ec2:DescribeVpcEndpointServices",
                        "iam:CreateServiceLinkedRole",
                        "kms:CreateCustomKeyStore",
                    ],
                    "Resource": ["*"],
                },
            ],
        }
        print(json.dumps(output, indent=4))
        self.maxDiff = None
        self.assertDictEqual(output, desired_output)

    #     def test_add_crud_with_wildcard(self):
    #         cfg = {
    #             "mode": "crud",
    #             "name": "RoleNameWithCRUD",
    #             "permissions-management": ["arn:aws:s3:::example-org-s3-access-logs"],
    #             "wildcard-only": {
    #                 "single-actions": [
    #                     # The first three are legitimately wildcard only.
    #                     # Verify with `policy_sentry query action-table --service secretsmanager --wildcard-only`
    #                     "ram:enablesharingwithawsorganization",
    #                     "ram:getresourcepolicies",
    #                     "secretsmanager:createsecret",
    #                     # This last one can be "secret" ARN type OR wildcard. We want to prevent people from
    #                     # bypassing this mechanism, while allowing them to explicitly
    #                     # request specific privs that require wildcard mode. This next value -
    #                     # secretsmanager:putsecretvalue - is an example of someone trying to beat the tool.
    #                     "secretsmanager:putsecretvalue",
    #                 ],
    #             }
    #         }
    #         sid_group = SidGroup()
    #         output = sid_group.process_template(cfg)
    #         desired_output = {
    #             "Version": "2012-10-17",
    #             "Statement": [
    #                 {
    #                     "Sid": "MultMultNone",
    #                     "Effect": "Allow",
    #                     "Action": [
    #                         "ram:EnableSharingWithAwsOrganization",
    #                         "ram:GetResourcePolicies",
    #                     ],
    #                     "Resource": ["*"],
    #                 },
    #                 {
    #                     "Sid": "S3PermissionsmanagementBucket",
    #                     "Effect": "Allow",
    #                     "Action": [
    #                         "s3:DeleteBucketPolicy",
    #                         "s3:PutBucketAcl",
    #                         "s3:PutBucketPolicy",
    #                         "s3:PutBucketPublicAccessBlock",
    #                     ],
    #                     "Resource": ["arn:aws:s3:::example-org-s3-access-logs"],
    #                 },
    #             ],
    #         }
    #         self.maxDiff = None
    #         # print(json.dumps(output, indent=4))
    #         self.assertDictEqual(output, desired_output)

    def test_sid_group_override(self):
        sid_group = SidGroup()
        output = sid_group.process_template(crud_with_override_template_cfg)
        self.maxDiff = None
        desired_output = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "MultMultNone",
                    "Effect": "Allow",
                    "Action": ["ram:GetResourcePolicies"],
                    "Resource": ["*"],
                },
                {
                    "Sid": "S3PermissionsmanagementBucket",
                    "Effect": "Allow",
                    "Action": [
                        "s3:DeleteBucketPolicy",
                        "s3:PutBucketAcl",
                        "s3:PutBucketPolicy",
                        "s3:PutBucketPublicAccessBlock",
                    ],
                    "Resource": ["arn:aws:s3:::example-org-s3-access-logs"],
                },
                {
                    "Sid": "SkipResourceConstraints",
                    "Effect": "Allow",
                    "Action": [
                        "ssm:GetParameter",
                        "ssm:GetParameters",
                        "ssm:GetParametersByPath",
                    ],
                    "Resource": ["*"],
                },
            ],
        }
        # print(json.dumps(output, indent=4))
        expected_statement_ids = [
            "MultMultNone",
            "S3PermissionsmanagementBucket",
            "SkipResourceConstraints",
        ]
        for statement in output.get("Statement"):
            self.assertTrue(statement.get("Sid") in expected_statement_ids)
        # self.assertDictEqual(output, desired_output)

    def test_exclude_actions_from_crud_output(self):
        sid_group = SidGroup()
        crud_with_exclude_actions = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                os.path.pardir,
                os.path.pardir,
                "examples",
                "yml",
                "crud-with-exclude-actions.yml",
            )
        )
        with open(crud_with_exclude_actions) as this_yaml_file:
            crud_with_exclude_actions_cfg = yaml.safe_load(this_yaml_file)
        sid_group.process_template(crud_with_exclude_actions_cfg)
        result = sid_group.get_rendered_policy(crud_with_exclude_actions_cfg)
        expected_statement_ids = ["MultMultNone", "KmsWriteKey"]
        expected_result = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "KmsWriteKey",
                    "Effect": "Allow",
                    "Action": [
                        "kms:CancelKeyDeletion",
                        "kms:CreateAlias",
                        "kms:Decrypt",
                        "kms:EnableKey",
                        "kms:EnableKeyRotation",
                        "kms:Encrypt",
                        "kms:GenerateDataKey",
                        "kms:GenerateDataKeyPair",
                        "kms:GenerateDataKeyPairWithoutPlaintext",
                        "kms:GenerateDataKeyWithoutPlaintext",
                        "kms:ImportKeyMaterial",
                        "kms:ReEncryptFrom",
                        "kms:ReEncryptTo",
                        "kms:Sign",
                        "kms:UpdateAlias",
                        "kms:UpdateKeyDescription",
                        "kms:Verify",
                    ],
                    "Resource": ["arn:aws:kms:us-east-1:123456789012:key/aaaa-bbbb-cccc"],
                }
            ],
        }
        # print(json.dumps(result, indent=4))
        for statement in result.get("Statement"):
            self.assertTrue(statement.get("Sid") in expected_statement_ids)
        # self.assertDictEqual(result, expected_result)

        crud_with_exclude_actions_cfg = {
            "mode": "crud",
            "write": ["arn:aws:kms:us-east-1:123456789012:key/aaaa-bbbb-cccc"],
            # This is really only because KMS is a special case.
            "exclude-actions": [
                "kms:Delete*",
                "kms:Disable*",
                "kms:Enable*",
                "kms:Generate*",
                "kms:Cancel*",
                "kms:Create*",
                "kms:Import*",
                "kms:ReEncrypt*",
                "kms:Sign*",
                "kms:Schedule*",
                "kms:Update*",
                "kms:Verify*",
            ],
        }
        sid_group.process_template(crud_with_exclude_actions_cfg)
        results = sid_group.get_rendered_policy()
        # print(json.dumps(results, indent=4))

        for statement in results.get("Statement"):
            self.assertTrue(statement.get("Sid") in expected_statement_ids)
        expected_result = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "KmsWriteKey",
                    "Effect": "Allow",
                    "Action": ["kms:Decrypt", "kms:Encrypt"],
                    "Resource": ["arn:aws:kms:us-east-1:123456789012:key/aaaa-bbbb-cccc"],
                }
            ],
        }
        expected_statement_ids = ["KmsWriteKey", "MultMultNone"]
        for statement in result.get("Statement"):
            self.assertTrue(statement.get("Sid") in expected_statement_ids)
        print(json.dumps(results, indent=4))
        self.maxDiff = None
        # self.assertDictEqual(results, expected_result)

    def test_exclude_actions_empty_sid_from_crud_output(self):
        sid_group = SidGroup()
        crud_with_exclude_actions_empty_sid = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                os.path.pardir,
                os.path.pardir,
                "examples",
                "yml",
                "crud-with-exclude-actions-empty-sid.yml",
            )
        )

        with open(crud_with_exclude_actions_empty_sid) as this_yaml_file:
            crud_with_exclude_actions_empty_sid_cfg = yaml.safe_load(this_yaml_file)
        # crud_with_exclude_actions_empty_sid_cfg = {
        #     "mode": "crud",
        #     "write": [
        #         "arn:aws:s3:::test"
        #     ],
        #     "exclude-actions": [
        #         "iam:Pass*"
        #     ]
        # }

        # print(json.dumps(crud_with_exclude_actions_empty_sid_cfg, indent=4))
        sid_group.process_template(crud_with_exclude_actions_empty_sid_cfg)
        result = sid_group.get_rendered_policy(crud_with_exclude_actions_empty_sid_cfg)
        # print(json.dumps(result, indent=4))
        expected_result = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "S3WriteBucket",
                    "Effect": "Allow",
                    "Action": [
                        "s3:CreateBucket",
                        "s3:DeleteBucket",
                        "s3:DeleteBucketWebsite",
                        "s3:PutAccelerateConfiguration",
                        "s3:PutAnalyticsConfiguration",
                        "s3:PutBucketCORS",
                        "s3:PutBucketLogging",
                        "s3:PutBucketNotification",
                        "s3:PutBucketObjectLockConfiguration",
                        "s3:PutBucketOwnershipControls",
                        "s3:PutBucketRequestPayment",
                        "s3:PutBucketVersioning",
                        "s3:PutBucketWebsite",
                        "s3:PutEncryptionConfiguration",
                        "s3:PutIntelligentTieringConfiguration",
                        "s3:PutInventoryConfiguration",
                        "s3:PutLifecycleConfiguration",
                        "s3:PutMetricsConfiguration",
                        "s3:PutReplicationConfiguration",
                    ],
                    "Resource": ["arn:aws:s3:::test"],
                }
            ],
        }
        self.maxDiff = None
        print(json.dumps(result, indent=4))
        actual_actions = next(
            (statement for statement in expected_result["Statement"] if statement["Sid"] == "S3WriteBucket"), None
        )
        expected_actions = expected_result["Statement"][0]["Action"]
        for action in expected_actions:
            self.assertIn(action, actual_actions["Action"])
        # self.assertDictEqual(result, expected_result)

    def test_write_template_with_sts_actions(self):
        cfg = {
            "mode": "crud",
            "name": "RoleNameWithCRUD",
            "sts": {"assume-role-with-web-identity": ["arn:aws:iam::123456789012:role/demo"]},
            "list": ["arn:aws:secretsmanager:us-east-1:123456789012:secret:anothersecret"],
        }
        sid_group = SidGroup()
        rendered_policy = sid_group.process_template(cfg)
        desired_output = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "AssumeRoleWithWebIdentity",
                    "Effect": "Allow",
                    "Action": ["sts:AssumeRoleWithWebIdentity"],
                    "Resource": ["arn:aws:iam::123456789012:role/demo"],
                }
            ],
        }

        # print(json.dumps(rendered_policy, indent=4))
        self.assertEqual(rendered_policy, desired_output)



"""
Given a Policy Sentry YML template, write a least-privilege IAM Policy in CRUD mode or Actions mode.
"""

from __future__ import annotations

import json
import logging
import sys
from typing import TYPE_CHECKING, Any

import click
import yaml
from click import Context

from policy_sentry import set_stream_logger
from policy_sentry.util.file import read_yaml_file
from policy_sentry.writing.sid_group import SidGroup

if TYPE_CHECKING:
    from pathlib import Path

logger = logging.getLogger(__name__)


# adapted from
# https://stackoverflow.com/questions/40753999/python-click-make-option-value-optional


class RegisterLengthOption(click.Option):
    """Mark this option as getting a _length option"""

    register_length = True


class RegisterLengthOptionHelp(click.Option):
    """Translate help for the hidden _length suffix"""

    def get_help_record(self, ctx: Context) -> tuple[str, str] | None:
        help_text = super().get_help_record(ctx)
        if help_text is None:
            return None

        return (help_text[0].replace("_length ", " "), *help_text[1:])


class RegisterMinimizeLengthCommand(click.Command):
    """Translate any opt= to opt_length= as needed"""

    def parse_args(self, ctx: Context, args: list[str]) -> list[str]:
        options = [o for o in ctx.command.params if getattr(o, "register_length", None)]
        prefixes = {p for p in sum((o.opts for o in options), []) if p.startswith("--")}  # noqa: RUF017
        for i, a in enumerate(args):
            a_tuple = a.split("=")
            if a_tuple[0] in prefixes:
                if len(a_tuple) > 1:
                    args[i] = a_tuple[0]
                    args.insert(i + 1, a_tuple[0] + "_length=" + a_tuple[1])  # noqa: B909
                else:
                    # check if next argument is naked
                    if len(args) > i + 1 and not args[i + 1].startswith("--"):
                        value = args[i + 1]
                        args[i + 1] = a_tuple[0] + "_length=" + value  # noqa: B909
        return super().parse_args(ctx, args)


@click.command(
    cls=RegisterMinimizeLengthCommand,
    short_help="Write least-privilege IAM policies, restricting all actions to resource ARNs.",
)
# pylint: disable=duplicate-code
@click.option(
    "--input-file",
    "-i",
    type=str,
    help="Path of the YAML File used for generating policies",
)
@click.option(
    "--minimize",
    "-m",
    cls=RegisterLengthOption,
    is_flag=True,
    required=False,
    default=False,
    help="Minimize the resulting statement with *safe* usage of wildcards to reduce policy length.",
)
@click.option(
    "--minimize_length",
    cls=RegisterLengthOptionHelp,
    help="Sets the minimum character length for minimization. Defaults to zero.",
    type=click.IntRange(0),
    required=False,
    default=0,
)
@click.option(
    "--fmt",
    type=click.Choice(["yaml", "json", "terraform"]),
    default="json",
    required=False,
    help='Format output as YAML or JSON. Defaults to "json"',
)
@click.option(
    "--verbose",
    "-v",
    type=click.Choice(["critical", "error", "warning", "info", "debug"], case_sensitive=False),
)
def write_policy(
    input_file: str | Path,
    minimize: int,
    minimize_length: int,
    fmt: str,
    verbose: str | None,
) -> None:
    """
    Write least-privilege IAM policies, restricting all actions to resource ARNs.
    """
    if verbose:
        log_level = getattr(logging, verbose.upper())
        set_stream_logger(level=log_level)

    if input_file:
        cfg = read_yaml_file(input_file)
    else:
        try:
            cfg = yaml.safe_load(sys.stdin)
        except yaml.YAMLError as exc:
            logger.critical(exc)
            sys.exit()

    min_length = None
    if minimize:
        min_length = minimize_length
    policy = write_policy_with_template(cfg, min_length)

    if fmt == "yaml":
        policy_str = yaml.dump(policy, sort_keys=False)
    else:
        indent = 4 if fmt == "json" else None
        policy_str = json.dumps(policy, indent=indent)
        if fmt == "terraform":
            obj = {"policy": policy_str}
            policy_str = json.dumps(obj)
    print(policy_str)


def write_policy_with_template(cfg: dict[str, Any], minimize: int | None = None) -> dict[str, Any]:
    """
    This function is called by write-policy so the config can be passed in as a dict without running into a Click-related error. Use this function, rather than the write-policy function, if you are using Policy Sentry as a python library.

    Arguments:
        cfg: The loaded YAML as a dict. Must follow Policy Sentry dictated format.
        minimize: Minimize the resulting statement with *safe* usage of wildcards to reduce policy length. Set this to the character length you want - for example, 0, or 4. Defaults to none.
    Returns:
        Dictionary: The JSON policy
    """
    sid_group = SidGroup()
    return sid_group.process_template(cfg, minimize)



"""Functions to expand wilcard actions into a full list of actions."""

from __future__ import annotations

import copy
import fnmatch
import logging
from typing import Any

from policy_sentry.querying.actions import get_actions_for_service
from policy_sentry.querying.all import get_all_actions
from policy_sentry.util.policy_files import get_actions_from_statement

logger = logging.getLogger(__name__)


def expand(action: str | list[str]) -> list[str]:
    """
    expand the action wildcards into a full action

    Arguments:
        action: An action in the form with a wildcard - like s3:Get*, or s3:L*
    Returns:
        List: A list of all the expanded actions (like actions matching s3:Get*)
    """

    if isinstance(action, list):
        expanded_actions = []
        for item in action:
            expanded_actions.extend(expand(item))
        return expanded_actions

    if action == "*":
        return list(get_all_actions())
    if "*" in action:
        service_prefix = action.split(":", maxsplit=1)[0]
        service_actions = get_actions_for_service(service_prefix=service_prefix)
        expanded = [
            expanded_action
            for expanded_action in service_actions
            if fnmatch.fnmatchcase(expanded_action.lower(), action.lower())
        ]

        # if we get a wildcard for a tech we've never heard of, just return the
        # wildcard
        if not expanded:
            logger.debug(
                "ERROR: The action %s references a wildcard for an unknown resource.",
                action,
            )
            return [action]

        return expanded
    return [action]


def determine_actions_to_expand(action_list: list[str]) -> list[str]:
    """
    Determine if an action needs to get expanded from its wildcard

    Arguments:
        action_list: A list of actions
    Returns:
        List: A list of actions
    """
    new_action_list = []
    for action in action_list:
        if "*" in action:
            expanded_action = expand(action)
            new_action_list.extend(expanded_action)
        else:
            # If there is no wildcard, copy that action name over to the new_action_list
            new_action_list.append(action)
    new_action_list.sort()
    return new_action_list


def get_expanded_policy(policy: dict[str, Any]) -> dict[str, Any]:
    """
    Given a policy, expand the * Actions in IAM policy files to improve readability

    Arguments:
        policy: dictionary containing valid AWS IAM Policy
    Returns:
        Dictionary: the policy that has the `*` expanded
    """
    modified_policy = copy.deepcopy(policy)

    modified_statement = modified_policy["Statement"]
    if isinstance(modified_statement, dict):
        requested_actions = get_actions_from_statement(modified_statement)
        expanded_actions = determine_actions_to_expand(requested_actions)
        if "NotAction" in modified_statement:
            if isinstance(modified_statement["NotAction"], list):
                modified_statement["NotAction"] = expanded_actions
            elif isinstance(modified_statement["NotAction"], str):
                modified_statement["NotAction"] = []
            logger.warning(
                "NotAction is in the statement. Policy Sentry will expand any wildcard actions "
                "that are in the NotAction list, but it will not factor the NotAction actions into any analysis about "
                "whether or not the actions are allowed by the policy. "
                "If you are concerned about this, please review this specific policy manually."
            )
        elif "Action" in modified_statement:
            modified_statement["Action"] = expanded_actions
    # Otherwise it will be a list of Sids
    elif isinstance(modified_statement, list):
        for statement in modified_statement:
            requested_actions = get_actions_from_statement(statement)
            expanded_actions = determine_actions_to_expand(requested_actions)
            if "NotAction" in statement:
                if isinstance(statement["NotAction"], list):
                    statement["NotAction"] = expanded_actions
                elif isinstance(statement["NotAction"], str):
                    statement["NotAction"] = []
                logger.warning(
                    "NotAction is in the statement. Policy Sentry will expand any wildcard actions "
                    "that are in the NotAction list, but it will not factor the NotAction actions into any analysis "
                    "about whether or not the actions are allowed by the policy."
                    "If you are concerned about this, please review this specific policy manually."
                )
            elif "Action" in statement:
                statement["Action"] = expanded_actions
    else:
        logger.critical("Unknown error: The 'Statement' is neither a dict nor a list")
    return modified_policy



import unittest
from policy_sentry.util.arns import does_arn_match, ARN, parse_arn

# "Does Arn Match" tests
# See docs for this list: # https://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html#genref-arns
# They removed some of the cases. The old version that gave this comprehensive list of resource ARN formats is here:
# https://web.archive.org/web/20190903192015/https://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html
# Case 1: arn:partition:service:region:account-id:resource
# Case 2: arn:partition:service:region:account-id:resourcetype/resource
# Case 3: arn:partition:service:region:account-id:resourcetype/resource/qualifier
# Case 4: arn:partition:service:region:account-id:resourcetype/resource:qualifier
# Case 5: arn:partition:service:region:account-id:resourcetype:resource
# Case 6: arn:partition:service:region:account-id:resourcetype:resource:qualifier
# Case 7: arn:partition:service:region:account-id:/greengrass/definition/devices/${DeviceDefinitionId}/versions/${VersionId}


class ArnsTestCase(unittest.TestCase):
    def test_does_arn_match_case_bucket(self):
        # Case 1: arn:partition:service:region:account-id:resource
        arn_to_test = "arn:aws:s3:::bucket_name"
        arn_in_database = "arn:${Partition}:s3:::${BucketName}"
        this_arn = ARN(arn_to_test)
        self.assertTrue(this_arn.same_resource_type(arn_in_database))

    def test_does_arn_match_case_1(self):
        # Case 1: arn:partition:service:region:account-id:resource
        arn_to_test = "arn:aws:codecommit:us-east-1:123456789012:MyDemoRepo"
        arn_in_database = "arn:${Partition}:codecommit:${Region}:${Account}:${RepositoryName}"
        this_arn = ARN(arn_to_test)
        self.assertTrue(this_arn.same_resource_type(arn_in_database))

    def test_does_arn_match_case_2(self):
        # Case 2: arn:partition:service:region:account-id:resourcetype/resource
        arn_to_test = "arn:aws:ssm:us-east-1:123456789012:parameter/test"
        arn_in_database = "arn:${Partition}:ssm:${Region}:${Account}:parameter/${FullyQualifiedParameterName}"
        this_arn = ARN(arn_to_test)
        self.assertTrue(this_arn.same_resource_type(arn_in_database))

    def test_does_arn_match_case_3(self):
        # Case 3: arn:partition:service:region:account-id:resourcetype/resource/qualifier
        arn_to_test = "arn:aws:kinesis:us-east-1:account-id:firehose/myfirehose/consumer/someconsumer:${ConsumerCreationTimpstamp}"
        arn_in_database = "arn:aws:kinesis:${Region}:${Account}:${StreamType}/${StreamName}/consumer/${ConsumerName}:${ConsumerCreationTimpstamp}"
        # https://docs.aws.amazon.com/kinesis/latest/APIReference/API_ConsumerDescription.html
        this_arn = ARN(arn_to_test)
        self.assertTrue(this_arn.same_resource_type(arn_in_database))

    def test_does_arn_match_case_4(self):
        # Case 4: arn:partition:service:region:account-id:resourcetype/resource:qualifier
        arn_to_test = "arn:aws:batch:region:account-id:job-definition/job-name:revision"
        arn_in_database = "arn:${Partition}:batch:${Region}:${Account}:job-definition/${JobDefinitionName}:${Revision}"
        this_arn = ARN(arn_to_test)
        self.assertTrue(this_arn.same_resource_type(arn_in_database))

    def test_does_arn_match_case_5(self):
        # Case 5: arn:partition:service:region:account-id:resourcetype:resource
        arn_to_test = "arn:aws:states:region:account-id:stateMachine:stateMachineName"
        arn_in_database = "arn:${Partition}:states:${Region}:${Account}:stateMachine:${StateMachineName}"
        this_arn = ARN(arn_to_test)
        self.assertTrue(this_arn.same_resource_type(arn_in_database))

    def test_does_arn_match_case_6(self):
        # Case 6: arn:partition:service:region:account-id:resourcetype:resource:qualifier
        arn_to_test = "arn:aws:states:region:account-id:execution:stateMachineName:executionName"
        arn_in_database = "arn:${Partition}:states:${Region}:${Account}:execution:${StateMachineName}:${ExecutionId}"
        this_arn = ARN(arn_to_test)
        self.assertTrue(this_arn.same_resource_type(arn_in_database))

    def test_does_arn_match_case_greengrass(self):
        # Undocumented case: AWS Greengrass: arn:aws:greengrass:${Region}:${Account}:/greengrass/definition/devices/${DeviceDefinitionId}/versions/${VersionId}
        arn_to_test = "arn:aws:greengrass:${Region}:${Account}:/greengrass/definition/devices/1234567/versions/1"
        arn_in_database = "arn:aws:greengrass:${Region}:${Account}:/greengrass/definition/devices/${DeviceDefinitionId}/versions/${VersionId}"
        this_arn = ARN(arn_to_test)
        self.assertTrue(this_arn.same_resource_type(arn_in_database))

    def test_does_arn_match_rds(self):
        arns_to_test = [
            "arn:${Partition}:rds:${Region}:${Account}:cluster:${DbClusterInstanceName}",
            "arn:${Partition}:rds:${Region}:${Account}:cluster-endpoint:${DbClusterEndpoint}",
            "arn:${Partition}:rds:${Region}:${Account}:cluster-pg:${ClusterParameterGroupName}",
            "arn:${Partition}:rds:${Region}:${Account}:cluster-snapshot:${ClusterSnapshotName}",
            "arn:${Partition}:rds:${Region}:${Account}:es:${SubscriptionName}",
            "arn:${Partition}:rds:${Account}:global-cluster:${GlobalCluster}",
            "arn:${Partition}:rds:${Region}:${Account}:og:${OptionGroupName}",
            "arn:${Partition}:rds:${Region}:${Account}:pg:${ParameterGroupName}",
            "arn:${Partition}:rds:${Region}:${Account}:db-proxy:${DbProxyId}",
            "arn:${Partition}:rds:${Region}:${Account}:ri:${ReservedDbInstanceName}",
            "arn:${Partition}:rds:${Region}:${Account}:secgrp:${SecurityGroupName}",
            "arn:${Partition}:rds:${Region}:${Account}:snapshot:${SnapshotName}",
            "arn:${Partition}:rds:${Region}:${Account}:subgrp:${SubnetGroupName}",
            "arn:${Partition}:rds:${Region}:${Account}:target:${TargetId}",
            "arn:${Partition}:rds:${Region}:${Account}:target-group:${TargetGroupId}",
        ]
        arn_in_database = "arn:${Partition}:rds:${Region}:${Account}:db:${DbInstanceName}"
        for arn in arns_to_test:
            this_arn = ARN(arn)
            self.assertFalse(this_arn.same_resource_type(arn_in_database))

        arn_to_test = "arn:${Partition}:rds:${Region}:${Account}:cluster:${DbClusterInstanceName}"
        this_arn = ARN(arn_to_test)
        self.assertFalse(this_arn.same_resource_type(arn_in_database))

        arn_to_test = "arn:${Partition}:rds:${Region}:${Account}:db:${DbInstanceName}"
        arn_in_database = "arn:${Partition}:rds:${Region}:${Account}:db:${DbInstanceName}"
        this_arn = ARN(arn_to_test)
        self.assertTrue(this_arn.same_resource_type(arn_in_database))

    def test_does_arn_match_resource_wildcard(self):
        arn_to_test = "arn:${Partition}:rds:${Region}:${Account}:*:*"
        arn_in_database = "arn:${Partition}:rds:${Region}:${Account}:db:${DbInstanceName}"
        this_arn = ARN(arn_to_test)
        self.assertTrue(this_arn.same_resource_type(arn_in_database))

        # Make sure wrong service yields False
        arn_to_test = "arn:${Partition}:s3:${Region}:${Account}:*:*"
        arn_in_database = "arn:${Partition}:rds:${Region}:${Account}:db:${DbInstanceName}"
        this_arn = ARN(arn_to_test)
        self.assertFalse(this_arn.same_resource_type(arn_in_database))

    def test_dynamodb_arn_matching_gh_215(self):
        """test_dynamodb_arn_matching_gh_215: Validate fix for DynamoDB arn mismatch in GitHub issue #215"""
        index = "arn:${Partition}:dynamodb:${Region}:${Account}:table/${TableName}/index/${IndexName}"
        stream = "arn:${Partition}:dynamodb:${Region}:${Account}:table/${TableName}/stream/${StreamLabel}"
        table = "arn:${Partition}:dynamodb:${Region}:${Account}:table/${TableName}"
        backup = "arn:${Partition}:dynamodb:${Region}:${Account}:table/${TableName}/backup/${BackupName}"
        global_table = "arn:${Partition}:dynamodb::${Account}:global-table/${GlobalTableName}"

        this_arn = ARN("arn:aws:dynamodb:us-east-1:123456789123:table/mytable")
        self.assertTrue(this_arn.same_resource_type(table))
        result = this_arn.same_resource_type(index)
        self.assertFalse(result)
        result = this_arn.same_resource_type(stream)
        self.assertFalse(result)
        result = this_arn.same_resource_type(backup)
        self.assertFalse(result)
        result = this_arn.same_resource_type(global_table)
        self.assertFalse(result)

        this_arn = "arn:aws:dynamodb:us-east-1:123456789123:table/mytable"
        self.assertTrue(does_arn_match(this_arn, table))
        self.assertFalse(does_arn_match(this_arn, index))
        self.assertFalse(does_arn_match(this_arn, stream))
        self.assertFalse(does_arn_match(this_arn, backup))
        self.assertFalse(does_arn_match(this_arn, global_table))

    def test_parse_arn(self):
        """util.arns.parse_arn: Ensure that invalid ARNs raise a proper exception message"""
        with self.assertRaises(Exception):
            parse_arn("aaa")


class ArnPathTestCase(unittest.TestCase):
    # When paths are used
    def test_ssm_paths(self):
        parameter_1 = ARN("arn:aws:ssm:::parameter/dev/foo/bar*")
        parameter_2 = "arn:aws:ssm:::parameter/dev"
        print(parameter_1.same_resource_type(parameter_2))
        self.assertTrue(parameter_1.same_resource_type(parameter_2))

    # When confusing ARNs that look like paths but are not actually paths are used
    def test_dynamo_db_non_paths(self):
        backup_arn = "arn:aws:dynamodb:us-east-1:123456789123:table/mytable/backup/mybackup"
        backup_raw_arn = "arn:${Partition}:dynamodb:${Region}:${Account}:table/${TableName}/backup/${BackupName}"

        table_arn = "arn:aws:dynamodb:us-east-1:123456789123:table/mytable"
        table_raw_arn = "arn:${Partition}:dynamodb:${Region}:${Account}:table/${TableName}"

        parameter_arn_with_path = "arn:aws:ssm:::parameter/dev/foo/bar*"
        parameter_arn_without_path = "arn:aws:ssm:::parameter/dev"
        parameter_raw_arn = "arn:${Partition}:ssm:${Region}:${Account}:parameter/${FullyQualifiedParameterName}"

        s3_object_with_path = "arn:aws:s3:::foo/bar/baz"
        s3_object_without_path = "arn:aws:s3:::foo/bar"

        s3_object_raw_arn = "arn:${Partition}:s3:::${BucketName}/${ObjectName}"
        s3_bucket_raw_arn = "arn:${Partition}:s3:::${BucketName}"

        ecr_raw_arn = "arn:${Partition}:ecr:${Region}:${Account}:repository/${RepositoryName}"
        ecr_arn_with_path = "arn:aws:ecr:*:*:repository/foo/bar"
        ecr_arn_without_path = "arn:aws:ecr:*:*:repository/foo"

        self.assertTrue(does_arn_match(backup_arn, backup_raw_arn))
        self.assertTrue(does_arn_match(table_arn, table_raw_arn))
        self.assertFalse(does_arn_match(table_arn, backup_raw_arn))
        self.assertFalse(does_arn_match(backup_arn, table_raw_arn))

        self.assertTrue(does_arn_match(parameter_arn_with_path, parameter_raw_arn))
        self.assertTrue(does_arn_match(parameter_arn_without_path, parameter_raw_arn))

        self.assertTrue(does_arn_match(s3_object_without_path, s3_object_raw_arn))
        self.assertTrue(does_arn_match(s3_object_with_path, s3_object_raw_arn))
        self.assertFalse(does_arn_match(s3_object_with_path, s3_bucket_raw_arn))
        self.assertFalse(does_arn_match(s3_object_without_path, s3_bucket_raw_arn))

        self.assertTrue(does_arn_match(ecr_arn_with_path, ecr_raw_arn))
        self.assertTrue(does_arn_match(ecr_arn_without_path, ecr_raw_arn))
